import os
from types import SimpleNamespace
from typing import List, Union

import pytest
from pytest import fixture

import fbnconfig
from fbnconfig import instrument_events as ctt
from fbnconfig import property as prop
from fbnconfig.resource_abc import Ref, Resource
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)
base_url: str = "api/api/instrumenteventtypes"


@fixture()
def base_resources(setup_deployment):
    deployment_name = setup_deployment.name

    # Create the property definition first
    property_def = prop.DefinitionResource(
        id="property_definition_for_transaction_template_int_test",
        code="MyCurrencyProperty",
        data_type_id=prop.ResourceId(scope="system", code="string"),
        display_name="My Currency Property",
        domain=prop.Domain.Transaction,
        scope=f"{deployment_name}withProps"
    )

    transaction_template = [
        property_def,
        ctt.TransactionTemplateResource(
            id="transaction_template_int_test",
            description="transaction template int test description",
            scope=deployment_name,
            instrument_event_type="BondCouponEvent",
            instrument_type="Bond",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                    ),
                    transaction_property_map=[],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
            ]
        ),
        ctt.TransactionTemplateResource(
            id="transaction_template_int_test_with_props",
            description="transaction template int test description",
            scope=f"{deployment_name}withProps",
            instrument_event_type="BondCouponEvent",  # This needs to be changed to an enum
            instrument_type="Bond",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                    ),
                    transaction_property_map=[
                        ctt.TransactionPropertyMap(
                            property_key=prop.DefinitionRef(
                                id="property_definition_id",
                                domain=prop.Domain.Transaction,
                                scope=f"{deployment_name}withProps",
                                code="MyCurrencyProperty"
                            ),
                            value="{{BondCouponEvent.currency}}"
                        )
                    ],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None)
            ]
        )
    ]

    return transaction_template


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("transaction_template")
    print(f"\nRunning deployment for {deployment_name}")
    yield SimpleNamespace(name=deployment_name)
    try:
        instrument_event = "BondCouponEvent"
        instrument_type = "Bond"
        formatted_url = (
            f"{base_url}/{instrument_event}/transactiontemplates/"
            f"{instrument_type}/{deployment_name}"
        )
        client.delete(formatted_url)
    except Exception:
        print("Failed to delete transaction template")
        pass


def test_create(setup_deployment, base_resources):
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deployex(deployment, lusid_env, token)

    instrument_event = "BondCouponEvent"
    instrument_type = "Bond"
    formatted_url = (
        f"{base_url}/{instrument_event}/transactiontemplates/"
        f"{instrument_type}/{setup_deployment.name}"
    )

    response = client.get(formatted_url).json()

    component = response["componentTransactions"][0]
    transaction_field_map = component["transactionFieldMap"]

    assert response["description"] == "transaction template int test description"
    assert response["instrumentType"] == "Bond"
    assert response["instrumentEventType"] == "BondCouponEvent"
    assert response["scope"] == setup_deployment.name
    assert component["displayName"] == "Bond Income Override"
    assert component["condition"] == "{{eligibleBalance}} gt 200"
    assert transaction_field_map["instrument"] == "{{instrument}}"
    assert transaction_field_map["settlementDate"] == "{{BondCouponEvent.paymentDate}}"
    assert transaction_field_map["transactionCurrency"] == "{{BondCouponEvent.currency}}"
    assert transaction_field_map["transactionDate"] == "{{BondCouponEvent.exDate}}"
    assert (
        transaction_field_map["transactionId"] ==
        "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
    )
    assert transaction_field_map["type"] == "BondCoupon"
    assert transaction_field_map["units"] == "{{eligibleBalance}}"
    assert (
        transaction_field_map["transactionPrice"]["price"] ==
        "{{BondCouponEvent.couponPerUnit}}"
    )
    assert transaction_field_map["transactionPrice"]["type"] == "CashFlowPerUnit"
    assert transaction_field_map["exchangeRate"] == "1"
    assert (
        transaction_field_map["totalConsideration"]["currency"] ==
        "{{BondCouponEvent.currency}}"
    )
    assert (
        transaction_field_map["totalConsideration"]["amount"] ==
        "{{BondCouponEvent.couponAmount}}"
    )
    assert component["transactionPropertyMap"] == []
    assert "preserveTaxLotStructure" not in component
    assert "marketOpenTimeAdjustments" not in component

    formatted_url_with_props = (
        f"{base_url}/{instrument_event}/transactiontemplates/"
        f"{instrument_type}/{setup_deployment.name}withProps"
    )

    response_with_props = client.get(formatted_url_with_props).json()

    component_with_props = response_with_props["componentTransactions"][0]
    transaction_field_map_with_props = component_with_props["transactionFieldMap"]

    assert response_with_props["description"] == "transaction template int test description"
    assert response_with_props["instrumentType"] == "Bond"
    assert response_with_props["instrumentEventType"] == "BondCouponEvent"
    assert response_with_props["scope"] == f"{setup_deployment.name}withProps"
    assert component_with_props["displayName"] == "Bond Income Override"
    assert component_with_props["condition"] == "{{eligibleBalance}} gt 200"
    assert (
        component_with_props["transactionFieldMap"]["instrument"] ==
        "{{instrument}}"
    )
    assert (
        transaction_field_map_with_props["settlementDate"] ==
        "{{BondCouponEvent.paymentDate}}"
    )
    assert (
        transaction_field_map_with_props["transactionCurrency"] ==
        "{{BondCouponEvent.currency}}"
    )
    assert (
        transaction_field_map_with_props["transactionDate"] ==
        "{{BondCouponEvent.exDate}}"
    )
    assert (
        transaction_field_map_with_props["transactionId"] ==
        "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
    )
    assert transaction_field_map_with_props["type"] == "BondCoupon"
    assert transaction_field_map_with_props["units"] == "{{eligibleBalance}}"
    assert (
        transaction_field_map_with_props["transactionPrice"]["price"] ==
        "{{BondCouponEvent.couponPerUnit}}"
    )
    assert (
        transaction_field_map_with_props["transactionPrice"]["type"] ==
        "CashFlowPerUnit"
    )
    assert transaction_field_map_with_props["exchangeRate"] == "1"
    assert (
        transaction_field_map_with_props["totalConsideration"]["currency"] ==
        "{{BondCouponEvent.currency}}"
    )
    assert (
        transaction_field_map_with_props["totalConsideration"]["amount"] ==
        "{{BondCouponEvent.couponAmount}}"
    )
    assert component_with_props["transactionPropertyMap"] == [
        {
            "propertyKey": f"Transaction/{setup_deployment.name}withProps/MyCurrencyProperty",
            "value": "{{BondCouponEvent.currency}}"
        }
    ]
    assert "preserveTaxLotStructure" not in response["componentTransactions"][0]


def test_update(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deployex(deployment, lusid_env, token)

    updated_template = ctt.TransactionTemplateResource(
        id="transaction_template_int_test",
        description="transaction template int test description",
        scope=deployment_name,
        instrument_event_type="BondCouponEvent",  # This needs to be changed to an enum
        instrument_type="Bond",
        component_transactions=[
            ctt.ComponentTransactions(
            display_name="Bond Income Override",
            condition="{{eligibleBalance}} gt 300",
            transaction_field_map=ctt.TransactionFieldMap(
                instrument="{{instrument}}",
                settlement_date="{{BondCouponEvent.paymentDate}}",
                source="MyOtherTransactionTypeSource",
                transaction_currency="{{BondCouponEvent.currency}}",
                transaction_date="{{BondCouponEvent.exDate}}",
                transaction_id="Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}",
                type="BondCoupon",
                units="{{eligibleBalance}}",
                transaction_price=ctt.TransactionPriceAndType(
                    price="{{BondCouponEvent.couponPerUnit}}",
                    type="CashFlowPerUnit"
                ),
                exchange_rate="2",
                total_consideration=ctt.TransactionCurrencyAndAmount(
                    currency="{{BondCouponEvent.currency}}",
                    amount="{{BondCouponEvent.couponAmount}}"
                ),
                ),
            transaction_property_map=[],
            preserve_tax_lot_structure=None,
            market_open_time_adjustments=None)
        ]
    )

    updated_deployment = fbnconfig.Deployment(setup_deployment.name, [updated_template])
    updated_resource = fbnconfig.deployex(updated_deployment, lusid_env, token)
    updated_list = [res.change for res in updated_resource if res.type == "TransactionTemplateResource"]

    # The second resource is 'removed' as it contains a transactionPropertyMap
    assert updated_list == ["update", "remove"]

    updated_template_data = client.get(
        f"{base_url}/{updated_template.instrument_event_type}/transactiontemplates/"
        f"{updated_template.instrument_type}/{deployment_name}"
    ).json()

    assert updated_template_data["componentTransactions"][0]["condition"] == "{{eligibleBalance}} gt 300"
    assert (
        updated_template_data["componentTransactions"][0]["transactionFieldMap"]["source"] ==
        "MyOtherTransactionTypeSource"
    )
    assert (
        updated_template_data["componentTransactions"][0]["transactionFieldMap"]["exchangeRate"] ==
        "2"
    )


def test_removing_transaction_property_map(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deployex(deployment, lusid_env, token)

    updated_template = ctt.TransactionTemplateResource(
        id="transaction_template_int_test",
        description="transaction template int test description",
        scope=deployment_name,
        instrument_event_type="BondCouponEvent",  # This needs to be changed to an enum
        instrument_type="Bond",
        component_transactions=[
            ctt.ComponentTransactions(
            display_name="Bond Income Override",
            condition="{{eligibleBalance}} gt 300",
            transaction_field_map=ctt.TransactionFieldMap(
                instrument="{{instrument}}",
                settlement_date="{{BondCouponEvent.paymentDate}}",
                source="MyOtherTransactionTypeSource",
                transaction_currency="{{BondCouponEvent.currency}}",
                transaction_date="{{BondCouponEvent.exDate}}",
                transaction_id="Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}",
                type="BondCoupon",
                units="{{eligibleBalance}}",
                transaction_price=ctt.TransactionPriceAndType(
                    price="{{BondCouponEvent.couponPerUnit}}",
                    type="CashFlowPerUnit"
                ),
                exchange_rate="2",
                total_consideration=ctt.TransactionCurrencyAndAmount(
                    currency="{{BondCouponEvent.currency}}",
                    amount="{{BondCouponEvent.couponAmount}}"
                ),
                ),
            transaction_property_map=[],
            preserve_tax_lot_structure=None,
            market_open_time_adjustments=None)
        ]
    )

    updated_deployment = fbnconfig.Deployment(setup_deployment.name, [updated_template])
    updated_resource = fbnconfig.deployex(updated_deployment, lusid_env, token)
    updated_list = [res.change for res in updated_resource if res.type == "TransactionTemplateResource"]
    updated_definition_list = [res.change for res in updated_resource
                               if res.type == "DefinitionResource"]

    assert updated_list == ["update", "remove"]
    assert updated_definition_list == ["remove"]

    updated_template_data = client.get(
        f"{base_url}/{updated_template.instrument_event_type}/transactiontemplates/"
        f"{updated_template.instrument_type}/{deployment_name}"
    ).json()
    component = updated_template_data["componentTransactions"][0]
    transaction_field_map = component["transactionFieldMap"]

    assert component["condition"] == "{{eligibleBalance}} gt 300"
    assert component["transactionPropertyMap"] == []
    assert transaction_field_map["source"] == "MyOtherTransactionTypeSource"
    assert transaction_field_map["exchangeRate"] == "2"


def test_adding_transaction_property_map(setup_deployment, base_resources):
    deployment_name = setup_deployment.name

    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deployex(deployment, lusid_env, token)

    # Create the property definition for the updated template
    property_def = prop.DefinitionResource(
        id="property_definition",
        code="MyCurrencyProperty",
        data_type_id=prop.ResourceId(scope="system", code="string"),
        display_name="My Currency Property",
        domain=prop.Domain.Transaction,
        scope=deployment_name
    )

    updated_template: List[Union[Resource, Ref]] = [
        property_def,
        ctt.TransactionTemplateResource(
        id="transaction_template_int_test",
        description="transaction template int test description",
        scope=deployment_name,
        instrument_event_type="BondCouponEvent",
        instrument_type="Bond",
        component_transactions=[
            ctt.ComponentTransactions(
            display_name="Bond Income Override",
            condition="{{eligibleBalance}} gt 200",
            transaction_field_map=ctt.TransactionFieldMap(
                instrument="{{instrument}}",
                settlement_date="{{BondCouponEvent.paymentDate}}",
                source="MyTransactionTypeSource",
                transaction_currency="{{BondCouponEvent.currency}}",
                transaction_date="{{BondCouponEvent.exDate}}",
                transaction_id="Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}",
                type="BondCoupon",
                units="{{eligibleBalance}}",
                transaction_price=ctt.TransactionPriceAndType(
                    price="{{BondCouponEvent.couponPerUnit}}",
                    type="CashFlowPerUnit"
                ),
                exchange_rate="1",
                total_consideration=ctt.TransactionCurrencyAndAmount(
                    currency="{{BondCouponEvent.currency}}",
                    amount="{{BondCouponEvent.couponAmount}}"
                ),
                ),
            transaction_property_map=[
                ctt.TransactionPropertyMap(
                    property_key=prop.DefinitionRef(
                        id="property_definition_id",
                        domain=prop.Domain.Transaction,
                        scope=deployment_name,
                        code="MyCurrencyProperty"
                    ),
                    value="{{BondCouponEvent.currency}}"
                )
            ],
            preserve_tax_lot_structure=None,
            market_open_time_adjustments=None)
        ]),
        ctt.TransactionTemplateResource(
            id="transaction_template_int_test_with_props",
            description="transaction template int test description",
            scope=f"{deployment_name}withProps",
            instrument_event_type="BondCouponEvent",  # This needs to be changed to an enum
            instrument_type="Bond",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                    ),
                    transaction_property_map=[],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None)
        ])
    ]

    updated_deployment = fbnconfig.Deployment(setup_deployment.name, updated_template)
    updated_resource = fbnconfig.deployex(updated_deployment, lusid_env, token)
    updated_list = [res.change for res in updated_resource if res.type == "TransactionTemplateResource"]
    updated_definition_list = [res.change for res in updated_resource
                               if res.type == "DefinitionResource"]

    assert updated_list == ["update", "update"]
    assert updated_definition_list == ["create", "remove"]

    property_def_url = (
        f"api/api/propertydefinitions/{prop.Domain.Transaction}/"
        f"{deployment_name}/MyCurrencyProperty"
    )

    property_def = client.get(property_def_url).json()
    assert property_def["scope"] == deployment_name
    assert property_def["code"] == "MyCurrencyProperty"
    assert property_def["domain"] == prop.Domain.Transaction
    assert property_def["dataTypeId"] == {"scope": "system", "code": "string"}
    assert property_def["displayName"] == "My Currency Property"

    assert_dict_contains(property_def, {
        "scope": deployment_name,
        "code": "MyCurrencyProperty",
        "domain": prop.Domain.Transaction,
        "dataTypeId": {"scope": "system", "code": "string"},
        "displayName": "My Currency Property"
    })

    # Resource 1
    transaction_template_url = (
        f"{base_url}/BondCouponEvent/transactiontemplates/"
        f"Bond/{setup_deployment.name}"
    )

    response = client.get(transaction_template_url).json()
    assert_dict_contains(response, {
        "scope": deployment_name,
        "instrumentEventType": "BondCouponEvent",
        "instrumentType": "Bond"
    })

    assert response["componentTransactions"][0]["transactionPropertyMap"][0] == {
        "propertyKey": f"Transaction/{setup_deployment.name}/MyCurrencyProperty",
        "value": "{{BondCouponEvent.currency}}"
    }

    transaction_template_with_props_url = (
        f"{base_url}/BondCouponEvent/transactiontemplates/"
        f"Bond/{setup_deployment.name}withProps"
    )

    transaction_template_with_no_props = client.get(transaction_template_with_props_url).json()
    assert_dict_contains(transaction_template_with_no_props, {
        "scope": f"{deployment_name}withProps",
        "instrumentEventType": "BondCouponEvent",
        "instrumentType": "Bond"
    })
    assert transaction_template_with_no_props["componentTransactions"][0]["transactionPropertyMap"] == []


def test_no_change(setup_deployment, base_resources):
    # given we have deployed the base case
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deployex(deployment, lusid_env, token)
    # when we apply it again
    update = fbnconfig.deployex(deployment, lusid_env, token)
    # then there are no changes
    ref_list_changes = [a.change for a in update if a.type == "TransactionTemplateResource"]
    assert ref_list_changes == ["nochange", "nochange"]


@pytest.mark.asyncio
async def test_transaction_template_list_with_filter(setup_deployment):
    # GIVEN TransactionTemplate is deployed
    deployment_name = setup_deployment.name
    scope = f"{deployment_name}withProps"
    event_type = "BondCouponEvent"
    instrument_type = "Bond"
    prop_code = "MyCurrencyProperty"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await ctt.TransactionTemplateResource.list(
        client=client,
        query_params={
            "filter": (f"scope eq '{scope}'"
                       f" and instrumentType eq '{instrument_type}'"
                       f" and instrumentEventType eq '{event_type}'")
        },
        path_params={}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
                "keys": {
                    "instrumentEventType": event_type,
                    "instrumentType": instrument_type,
                    "scope": scope
                },
                "value": {
                    "id": f"transaction-template_{scope}_{instrument_type}_{event_type}".lower(),
                    "instrumentEventType": event_type,
                    "instrumentType": instrument_type,
                    "scope": scope,
                    "description": "transaction template int test description",
                    "componentTransactions": [
                        {
                            "condition": "{{eligibleBalance}} gt 200",
                            "displayName": "Bond Income Override",
                            "transactionFieldMap": {
                                "exchangeRate": "1",
                                "instrument": "{{instrument}}",
                                "settlementDate": "{{BondCouponEvent.paymentDate}}",
                                "source": "MyTransactionTypeSource",
                                "totalConsideration": {
                                    "amount": "{{BondCouponEvent.couponAmount}}",
                                    "currency": "{{BondCouponEvent.currency}}",
                                },
                                "transactionCurrency": "{{BondCouponEvent.currency}}",
                                "transactionDate": "{{BondCouponEvent.exDate}}",
                                "transactionId": (
                                    "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                                ),
                                "transactionPrice": {
                                    "price": "{{BondCouponEvent.couponPerUnit}}",
                                    "type": "CashFlowPerUnit",
                                },
                                "type": "BondCoupon",
                                "units": "{{eligibleBalance}}",
                            },
                            "transactionPropertyMap": [
                                {
                                    "propertyKey": {
                                        "$ref": f"definition_transaction_{scope}_{prop_code}".lower()
                                    },
                                    "value": "{{BondCouponEvent.currency}}"
                                }
                            ]
                        }
                    ]
                }
            }
        ]
    }
    assert result == expected_result


@pytest.mark.asyncio
async def test_transaction_template_fetch(setup_deployment):
    # GIVEN TransactionTemplate is deployed
    deployment_name = setup_deployment.name
    scope = f"{deployment_name}withProps"
    instrument_event_type = "BondCouponEvent"
    instrument_type = "Bond"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await ctt.TransactionTemplateResource.fetch(
        client=client,
        keys={
            "instrumentEventType": instrument_event_type,
            "instrumentType": instrument_type,
            "scope": scope
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"transaction-template_{scope}_{instrument_type}_{instrument_event_type}".lower(),
        "instrumentEventType": instrument_event_type,
        "instrumentType": instrument_type,
        "scope": scope,
        "description": "transaction template int test description",
        "componentTransactions": [
            {
                "condition": "{{eligibleBalance}} gt 200",
                "displayName": "Bond Income Override",
                "transactionFieldMap": {
                    "exchangeRate": "1",
                    "instrument": "{{instrument}}",
                    "settlementDate": "{{BondCouponEvent.paymentDate}}",
                    "source": "MyTransactionTypeSource",
                    "totalConsideration": {
                        "amount": "{{BondCouponEvent.couponAmount}}",
                        "currency": "{{BondCouponEvent.currency}}",
                    },
                    "transactionCurrency": "{{BondCouponEvent.currency}}",
                    "transactionDate": "{{BondCouponEvent.exDate}}",
                    "transactionId": "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}",
                    "transactionPrice": {
                        "price": "{{BondCouponEvent.couponPerUnit}}",
                        "type": "CashFlowPerUnit",
                    },
                    "type": "BondCoupon",
                    "units": "{{eligibleBalance}}",
                },
                "transactionPropertyMap": [
                    {
                        "propertyKey": {
                            "$ref": f"definition_transaction_{scope}_mycurrencyproperty".lower()
                        },
                        "value": "{{BondCouponEvent.currency}}"
                    }
                ]
            }
        ]
    }
    assert response == expected_response


def test_teardown(setup_deployment, base_resources):
    deployment_name = setup_deployment.name
    # given we have deployed the base case
    deployment = fbnconfig.Deployment(deployment_name, base_resources)
    fbnconfig.deployex(deployment, lusid_env, token)
    # when we remove all the resources
    empty = fbnconfig.Deployment(deployment_name, [])
    update = fbnconfig.deployex(empty, lusid_env, token)
    # then there are no changes
    ref_list_changes = [a.change for a in update if a.type == "TransactionTemplateResource"]
    assert ref_list_changes == ["remove", "remove"]


def assert_dict_contains(actual, expected):
    for key, value in expected.items():
        assert key in actual, f"Missing key: {key}"
        assert actual[key] == value, f"Key '{key}': expected {value}, got {actual[key]}"



import json
from hashlib import sha256
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig.corporate_action import CorporateActionSourceResource

TEST_BASE = "https://foo.lusid.com"


class DescribeCorporateActionSourceResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    def get_hash(self, obj: CorporateActionSourceResource) -> str:
        desired = obj.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        return sha256(sorted_desired.encode()).hexdigest()

    def test_create_minimal(self, respx_mock):
        respx_mock.post("/api/api/corporateactionsources").mock(
            return_value=httpx.Response(201, json={})
        )
        client = self.client
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source"
        )
        new_state = sut.create(client)
        post = respx_mock.calls[0]
        assert post.request.url == f"{TEST_BASE}/api/api/corporateactionsources"
        assert json.loads(post.request.content) == {
            "scope": "test",
            "code": "ca1",
            "displayName": "Test Corporate Action Source"
        }
        assert new_state == {"scope": "test", "code": "ca1", "content_hash": self.get_hash(sut)}

    def test_create_with_description_and_instrument_scopes(self, respx_mock):
        respx_mock.post("/api/api/corporateactionsources").mock(
            return_value=httpx.Response(201, json={})
        )
        client = self.client
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source",
            description="A test corporate action source",
            instrument_scopes=["default"]
        )
        new_state = sut.create(client)
        post = respx_mock.calls[0]
        assert json.loads(post.request.content) == {
            "scope": "test",
            "code": "ca1",
            "displayName": "Test Corporate Action Source",
            "description": "A test corporate action source",
            "instrumentScopes": ["default"]
        }
        assert new_state == {"scope": "test", "code": "ca1", "content_hash": self.get_hash(sut)}

    def test_instrument_scopes_validation_none_allowed(self):
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source",
            instrument_scopes=None
        )
        assert sut.instrument_scopes is None

    def test_instrument_scopes_validation_empty_list_allowed(self):
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source",
            instrument_scopes=[]
        )
        assert sut.instrument_scopes == []

    def test_instrument_scopes_validation_single_element_allowed(self):
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source",
            instrument_scopes=["default"]
        )
        assert sut.instrument_scopes == ["default"]

    def test_instrument_scopes_validation_multiple_elements_rejected(self):
        with pytest.raises(ValueError, match="instrument_scopes can have at most one element"):
            CorporateActionSourceResource(
                id="test_ca",
                scope="test",
                code="ca1",
                display_name="Test Corporate Action Source",
                instrument_scopes=["default", "test"]
            )

    def test_read(self, respx_mock):
        respx_mock.get("/api/api/corporateactionsources").mock(
            return_value=httpx.Response(200, json={
                "values": [{
                    "id": {"scope": "test", "code": "ca1"},
                    "displayName": "Test Corporate Action Source",
                    "instrumentScopes": ["default"]
                }]
            })
        )
        client = self.client
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source"
        )
        result = sut.read(client, SimpleNamespace(scope="test", code="ca1"))
        req = respx_mock.calls[0]
        assert req.request.url.path == "/api/api/corporateactionsources"
        assert req.request.url.query == b"filter=id.scope+eq+%27test%27+and+id.code+eq+%27ca1%27"
        assert result and result["id"]
        assert result["id"]["scope"] == "test"
        assert result["id"]["code"] == "ca1"

    def test_read_not_found(self, respx_mock):
        respx_mock.get("/api/api/corporateactionsources").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        client = self.client
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source"
        )
        result = sut.read(client, SimpleNamespace(scope="test", code="ca1"))
        assert result is None

    def test_delete(self, respx_mock):
        respx_mock.delete("/api/api/corporateactionsources/test/ca1").mock(
            return_value=httpx.Response(204, json={})
        )
        client = self.client
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source"
        )
        sut.delete(client, SimpleNamespace(scope="test", code="ca1"))
        req = respx_mock.calls[0]
        assert req.request.url == f"{TEST_BASE}/api/api/corporateactionsources/test/ca1"

    def test_update_no_change(self):
        client = self.client
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source"
        )
        remote_state = SimpleNamespace(scope="test", code="ca1", content_hash=self.get_hash(sut))
        new_state = sut.update(client, remote_state)
        assert not new_state

    def test_update_with_change(self, respx_mock):
        respx_mock.delete("/api/api/corporateactionsources/test/ca1").mock(
            return_value=httpx.Response(204, json={})
        )
        respx_mock.post("/api/api/corporateactionsources").mock(
            return_value=httpx.Response(201, json={})
        )
        client = self.client
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Updated Corporate Action Source",
            description="Updated description"
        )
        remote_state = SimpleNamespace(scope="test", code="ca1", content_hash="old_hash")
        new_state = sut.update(client, remote_state)
        delete_req = respx_mock.calls[0]
        assert delete_req.request.url == f"{TEST_BASE}/api/api/corporateactionsources/test/ca1"
        assert delete_req.request.method == "DELETE"
        create_req = respx_mock.calls[1]
        assert create_req.request.url == f"{TEST_BASE}/api/api/corporateactionsources"
        assert create_req.request.method == "POST"
        assert json.loads(create_req.request.content) == {
            "scope": "test",
            "code": "ca1",
            "displayName": "Updated Corporate Action Source",
            "description": "Updated description"
        }
        assert new_state == {"scope": "test", "code": "ca1", "content_hash": self.get_hash(sut)}

    def test_deps(self):
        sut = CorporateActionSourceResource(
            id="test_ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source"
        )
        assert sut.deps() == []

    def test_model_validation_from_dict_with_id_object(self):
        data = {
            "id": {"scope": "test", "code": "ca1"},
            "displayName": "Test Corporate Action Source",
            "description": "A test source",
            "instrumentScopes": ["default"]
        }
        sut = CorporateActionSourceResource.model_validate(data, context={"id": "test_ca"})
        assert sut.id == "test_ca"
        assert sut.scope == "test"
        assert sut.code == "ca1"
        assert sut.display_name == "Test Corporate Action Source"
        assert sut.description == "A test source"
        assert sut.instrument_scopes == ["default"]

    def test_dump_undump_with_id_context(self):
        sut = CorporateActionSourceResource(
            id="test-ca",
            scope="test",
            code="ca1",
            display_name="Test Corporate Action Source",
            description="A test corporate action source",
            instrument_scopes=["default"]
        )
        dumped = sut.model_dump(
            mode="json",
            by_alias=True,
            exclude_none=True,
            round_trip=True,
            context={"style": "dump"}
        )
        assert "id" not in dumped
        assert dumped == {
            "scope": "test",
            "code": "ca1",
            "displayName": "Test Corporate Action Source",
            "description": "A test corporate action source",
            "instrumentScopes": ["default"]
        }
        result = CorporateActionSourceResource.model_validate(
            dumped, context={"style": "dump", "id": "context-ca"}
        )
        assert result.id == "context-ca"
        assert result.scope == "test"
        assert result.code == "ca1"
        assert result.display_name == "Test Corporate Action Source"
        assert result.description == "A test corporate action source"
        assert result.instrument_scopes == ["default"]

    def test_from_api(self, testdata_loader):
        # GIVEN a CorporateActionSource resource
        input_data = testdata_loader(
            "corporate_action_source_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "corporate_action_source_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # WHEN the from_api method is called for that resource
        result = CorporateActionSourceResource.from_api(input_data)
        # THEN the response is validated successfully
        casr_obj = CorporateActionSourceResource.model_validate(result)
        assert casr_obj.id == "corporate-action-source_lusid-bloomberg_dl-per-security"
        # AND the response is as expected
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        # GIVEN a CorporateActionSource resource
        input_data = testdata_loader(
            "corporate_action_source_resource",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "corporate_action_source_resource",
            "list_expected_result"
        )
        respx_mock.get(CorporateActionSourceResource.list_endpoint()).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the list method is called for that resource
        async with httpx.AsyncClient(base_url=TEST_BASE) as client:
            result = await CorporateActionSourceResource.list(client, query_params={})
        assert result == expected_result
        # THEN the response is as expected
        assert respx_mock.calls.last.request.method == "GET"
        assert "/api/api/corporateactionsources" in str(
            respx_mock.calls.last.request.url
        )

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        # GIVEN a CorporateActionSource resource
        input_data = testdata_loader(
            "corporate_action_source_resource",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "corporate_action_source_resource",
            "list_expected_result"
        )["values"][0]["value"]
        keys = {
            "scope": "LUSID-Bloomberg",
            "code": "dl-per-security"
        }
        respx_mock.get(CorporateActionSourceResource.fetch_endpoint()).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the fetch method is called for that resource
        async with httpx.AsyncClient(base_url=TEST_BASE) as client:
            result = await CorporateActionSourceResource.fetch(client, keys=keys)
        # THEN the response is as expected
        assert result == expected_result



import os

import pytest

import fbnconfig
import tests.integration.identity as identity

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


def test_teardown():
    deployment_name = identity.configure(host_vars).id
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_roles_by_name(client, "robTest_role2")
    assert len(matches) == 0
    users = get_users_by_login(client, "robtest_jane3@robtest.com")
    assert len(users) == 0


@pytest.mark.skip(reason="This creates actual users; should be run explicitly")
def test_create():
    fbnconfig.deployex(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_roles_by_name(client, "robTest_role2")
    assert len(matches) == 1
    users = get_users_by_login(client, "robtest_jane3@robtest.com")
    assert len(users) == 1


def test_update():
    fbnconfig.deployex(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_roles_by_name(client, "robTest_role2")
    assert len(matches) == 1
    users = get_users_by_login(client, "robtest_jane3@robtest.com")
    assert len(users) == 1


def get_roles_by_name(client, name):
    get = client.request("get", "/identity/api/roles")
    get.raise_for_status()
    roles = get.json()
    return [role for role in roles if role["name"] == name]


def get_users_by_login(client, login):
    get = client.request("get", "/identity/api/users")
    get.raise_for_status()
    users = get.json()
    return [user for user in users if user["login"] == login]


@pytest.mark.asyncio
async def test_role_list():
    # GIVEN Role is deployed
    fbnconfig.deployex(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    code = "robTest_role2"
    # WHEN the fetch method is called for that resource
    response = await identity.identity.IdentityRoleResource.list(
        client=client,
        query_params={},
        path_params={}
    )
    role = [role for role in response["values"] if role["value"]["name"] == code]
    role_id = role[0]["value"]["roleId"]
    # THEN the response is as expected
    expected_role = [{
        "keys": {
            "roleId": role_id
        },
        "value": {
            "id": f"identity-role_{role_id}".lower(),
            "roleId": role_id,
            "code": code,
            "name": code,
            "scope": "default",
            "description": "scheduler-operators"
        }
    }]
    assert role == expected_role


@pytest.mark.asyncio
async def test_role_fetch():
    # GIVEN Role is deployed
    fbnconfig.deployex(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    roles_response = await client.request("get", "/identity/api/roles")
    roles = roles_response.json()
    code = "robTest_role2"
    role_id = [role for role in roles if role["name"] == code][0]["id"]
    # WHEN the fetch method is called for that resource
    response = await identity.identity.IdentityRoleResource.fetch(
        client=client,
        keys={
            "roleId": role_id
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"identity-role_{role_id}".lower(),
        "roleId": role_id,
        "code": code,
        "name": code,
        "scope": "default",
        "description": "scheduler-operators"
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_role_assignment_list():
    # GIVEN RoleAssignment is deployed
    fbnconfig.deployex(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    roles_response = await client.request("get", "/identity/api/roles")
    roles = roles_response.json()
    role_code = "robTest_role2"
    role_id = [role for role in roles if role["name"] == role_code][0]["id"]
    users_response = await client.request("get", "/identity/api/users")
    users = users_response.json()
    user_login = "robtest_jane3@robtest.com"
    user_id = [user for user in users if user["login"] == user_login][0]["id"]
    # WHEN the fetch method is called for that resource
    response = await identity.identity.RoleAssignment.list(
        client=client,
        query_params={},
        path_params={
            "roleId": role_id,
        }
    )
    # THEN the response is as expected
    expected_role = {
        "values": [
            {
                "keys": {
                    "roleId": role_id,
                    "userId": user_id
                },
                "value": {
                    "id": f"role-assignment_{role_id}_{user_id}".lower(),
                    "roleId": role_id,
                    "userId": user_id,
                    "role": {
                        "$ref": f"identity-role_{role_id}".lower()
                    },
                    "user": {
                        "$ref": f"user_{user_id}".lower()
                    }
                }
            }
        ]
    }
    assert response == expected_role


@pytest.mark.asyncio
async def test_user_list():
    # GIVEN User is deployed
    fbnconfig.deployex(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    login = "robtest_jane3@robtest.com"
    # WHEN the fetch method is called for that resource
    response = await identity.identity.UserResource.list(
        client=client,
        query_params={},
        path_params={}
    )
    user = [user for user in response["values"] if user["value"]["login"] == login]
    user_id = user[0]["keys"]["userId"]
    # THEN the response is as expected
    expected_role = [{
        "keys": {
            "userId": user_id
        },
        "value": {
            "id": f"user_{user_id}".lower(),
            "emailAddress": "robtest_jane@robert.byrnemail.org",
            "login": login,
            "firstName": "robTestJane",
            "lastName": "robTestSmith",
            "type": "Service",
            "secondEmailAddress": None
        }
    }]
    assert user == expected_role


@pytest.mark.asyncio
async def test_user_fetch():
    # GIVEN User is deployed
    fbnconfig.deployex(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    users_response = await client.request("get", "/identity/api/users")
    users = users_response.json()
    login = "robtest_jane3@robtest.com"
    user_id = [user for user in users if user["login"] == login][0]["id"]
    # WHEN the fetch method is called for that resource
    response = await identity.identity.UserResource.fetch(
        client=client,
        keys={
            "userId": user_id
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"user_{user_id}".lower(),
        "emailAddress": "robtest_jane@robert.byrnemail.org",
        "login": login,
        "firstName": "robTestJane",
        "lastName": "robTestSmith",
        "type": "Service",
        "secondEmailAddress": None
    }
    assert response == expected_response



import os
from types import SimpleNamespace

import pytest
from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
from fbnconfig import datatype, property
from tests.integration.generate_test_name import gen


@fixture(scope="module")
def lusid_env():
    if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
        raise (
            RuntimeError("FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
        )

    env = os.environ["LUSID_ENV"]
    token = os.environ["FBN_ACCESS_TOKEN"]
    return SimpleNamespace(env=env, token=token)


@fixture(scope="module")
def client(lusid_env):
    return fbnconfig.create_client(lusid_env.env, lusid_env.token)


@fixture(scope="module")
def deployment_name():
    return gen("property")


def resources(deployment_name):
    df = property.DefinitionRef(
        id="property/dfdomccy", domain=property.Domain.Holding, scope="default", code="DfDomCcy"
    )
    nominal = property.DefinitionRef(
        id="property/nominal", domain=property.Domain.Holding, scope="default", code="Nominal"
    )
    rating = property.DefinitionResource(
        id="rating",
        domain=property.Domain.Instrument,
        scope=deployment_name,
        code="rating",
        display_name="robtest rating ",
        data_type_id=datatype.DataTypeRef(id="default_str", scope="system", code="string"),
        constraint_style=property.ConstraintStyle.Collection,
        property_description="robTest property",
        life_time=property.LifeTime.Perpetual,
        collection_type=property.CollectionType.Array,
    )
    pd = property.DefinitionResource(
        id="pd1",
        domain=property.Domain.Instrument,
        scope=deployment_name,
        code="pd1",
        display_name="robtest pd ",
        data_type_id=property.ResourceId(scope="system", code="number"),
        constraint_style=property.ConstraintStyle.Property,
        property_description="robTest property",
        life_time=property.LifeTime.Perpetual,
        collection_type=None,
    )
    pv_nominal = property.DefinitionResource(
        id="derived",
        domain=property.Domain.Holding,
        data_type_id=property.ResourceId(scope="system", code="number"),
        scope=deployment_name,
        code="PVNominal",
        property_description="nominal_x_df",
        display_name="DF Nominal",
        derivation_formula=property.Formula("{x} * {y}", x=df, y=nominal),
        is_filterable=False,
    )
    more_derived = property.DefinitionResource(
        id="derived_more",
        domain=property.Domain.Holding,
        data_type_id=property.ResourceId(scope="system", code="number"),
        scope=deployment_name,
        code="more_derived",
        property_description="pd1 x df x nominal",
        display_name="DF Nominal pd1",
        derivation_formula=property.Formula("{x} * {y}", x=pv_nominal, y=pd),
        is_filterable=True,
    )
    return {r.id: r for r in [df, nominal, rating, pd, pv_nominal, more_derived]}


@fixture(scope="module")
def deployment(deployment_name, lusid_env):
    res = resources(deployment_name)
    print(f"\nRunning for deployment {deployment_name}...")
    yield fbnconfig.Deployment(deployment_name, list(res.values()))
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env.env, lusid_env.token)


def test_teardown(deployment, lusid_env, client):
    # create first
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    fbnconfig.deployex(fbnconfig.Deployment(deployment.id, []), lusid_env.env, lusid_env.token)
    with pytest.raises(HTTPStatusError) as error:
        client.get(f"/api/api/propertydefinitions/Holding/{deployment.id}/more_derived")
    assert error.value.response.status_code == 404


def test_create(deployment, lusid_env, client):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    search = client.get(f"/api/api/propertydefinitions/Holding/{deployment.id}/more_derived")
    assert search.status_code == 200


def test_update(deployment, lusid_env, client):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    search = client.get(f"/api/api/propertydefinitions/Holding/{deployment.id}/more_derived")
    assert search.status_code == 200


@pytest.mark.asyncio
async def test_list_with_filter(deployment, lusid_env):
    # GIVEN Definition is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    domain = property.Domain.Holding.value
    scope = deployment.id
    code = "more_derived"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN calling list with filter for the deployment name
    result = await property.DefinitionResource.list(
        client=client,
        query_params={"filter": f"code eq '{code}' and scope eq '{scope}'"}
    )
    # THEN the result has correct structure
    expected_result = {
        "values": [
            {
            "keys": {"code": code, "scope": scope, "domain": domain},
            "value":
            {
                "id": f"definition_{domain}_{scope}_{code}".lower(),
                "valueType": "Decimal",
                "displayName": "DF Nominal pd1",
                "dataTypeId": {
                    "$ref": "datatype_system_number"
                },
                "type": "Metric",
                "unitSchema": "NoUnits",
                "domain": domain,
                "scope": scope,
                "code": code,
                "lifeTime": "TimeVariant",
                "propertyDescription": "pd1 x df x nominal",
                "derivationFormula": {
                    "formula": "{p1} * {p2}",
                    "args": {
                        "p1": {
                            "$ref": f"definition_holding_{scope}_pvnominal".lower()
                        },
                        "p2": {
                            "$ref": f"definition_instrument_{scope}_pd1".lower()
                        }
                    }
                },
            }
            }
        ]
    }
    assert result["values"] == expected_result["values"]


@pytest.mark.asyncio
async def test_list_returns_next_page(deployment, lusid_env):
    # GIVEN Definition is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN calling list with limit of 2
    result_page = await property.DefinitionResource.list(
        client=client,
        query_params={"limit": 2}
    )
    # THEN the result has correct structure
    assert "values" in result_page and isinstance(result_page["values"], list)
    assert "nextPage" in result_page


@pytest.mark.asyncio
async def test_fetch(deployment, lusid_env):
    # GIVEN Definition is deployed
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    domain = property.Domain.Holding.value
    scope = deployment.id
    code = "more_derived"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await property.DefinitionResource.fetch(
        client=client,
        keys={
            "scope": scope,
            "code": code,
            "domain": domain,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"definition_{domain}_{scope}_{code}".lower(),
        "valueType": "Decimal",
        "displayName": "DF Nominal pd1",
        "dataTypeId": {
            "$ref": "datatype_system_number"
        },
        "type": "Metric",
        "unitSchema": "NoUnits",
        "domain": domain,
        "scope": scope,
        "code": code,
        "lifeTime": "TimeVariant",
        "propertyDescription": "pd1 x df x nominal",
        "derivationFormula": {
            "formula": "{p1} * {p2}",
            "args": {
                "p1": {
                    "$ref": f"definition_holding_{scope}_pvnominal".lower()
                },
                "p2": {
                    "$ref": f"definition_instrument_{scope}_pd1".lower()
                }
            }
        },
    }
    assert response == expected_response



import os
from types import SimpleNamespace

import pytest
from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
from fbnconfig import side_definition, transaction_type
from tests.integration.generate_test_name import gen


@fixture(scope="module")
def lusid_env():
    if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
        raise (
            RuntimeError("FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
        )

    env = os.environ["LUSID_ENV"]
    token = os.environ["FBN_ACCESS_TOKEN"]
    return SimpleNamespace(env=env, token=token)


@fixture(scope="module")
def client(lusid_env):
    return fbnconfig.create_client(lusid_env.env, lusid_env.token)


@fixture(scope="module")
def deployment_name():
    return gen("transaction_configuration")


def resources(deployment_name):
    test_side1 = side_definition.SideResource(
        id="side1",
        side="Side1",
        scope=deployment_name,
        security="Txn:LusidInstrumentId",
        currency="Txn:SettlementCurrency",
        rate="Txn:TradeToPortfolioRate",
        units="Txn:Units",
        amount="Txn:TotalConsideration",
        notional_amount="0",
    )

    test_side2 = side_definition.SideResource(
        id="side2",
        side="Side2",
        scope=deployment_name,
        security="Txn:SettlementCurrency",
        currency="Txn:SettlementCurrency",
        rate="SettledToPortfolioRate",
        units="Txn:TotalConsideration",
        amount="Txn:TotalConsideration",
        notional_amount="0",
        current_face="Txn:TotalConsideration",
    )

    test_buy_transaction_type = transaction_type.TransactionTypeResource(
        id="example",
        scope=deployment_name,
        source="default",
        aliases=[
            transaction_type.TransactionTypeAlias(
                type="Buy",
                description="Something",
                transaction_class="default",
                transaction_roles="LongLonger",
                is_default=False,
            ),
            transaction_type.TransactionTypeAlias(
                type="BY",
                description="Something but different",
                transaction_class="default",
                transaction_roles="LongLonger",
                is_default=False,
            ),
        ],
        movements=[
            transaction_type.TransactionTypeMovement(
                movement_types=transaction_type.MovementType.StockMovement,
                side=test_side1,
                direction=1,
                name="Stock Movement",
                movement_options=[transaction_type.MovementOption.DirectAdjustment],
            )
        ],
        calculations=[
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.TaxAmounts, side=test_side2
            ),
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.NotionalAmount
            ),
        ],
    )
    return {"side1": test_side1, "side2": test_side2, "type": test_buy_transaction_type}


@fixture(scope="module")
def deployment(deployment_name, lusid_env):
    res = resources(deployment_name)
    print(f"\nRunning for deployment {deployment_name}...")
    yield fbnconfig.Deployment(deployment_name, list(res.values()))
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env.env, lusid_env.token)


def test_teardown_side(deployment, lusid_env, client):
    # create first
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    fbnconfig.deployex(fbnconfig.Deployment(deployment.id, []), lusid_env.env, lusid_env.token)
    with pytest.raises(HTTPStatusError) as error:
        client.get(
            "/api/api/transactionconfiguration/sides/Side1", params={"scope": deployment.id}
        )
        assert error.value.response.status_code == 404


def test_create_side(deployment, lusid_env, client):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    search = client.get(
        "/api/api/transactionconfiguration/sides/Side1", params={"scope": deployment.id}
    )
    assert search.status_code == 200


def test_nochange_side(deployment, lusid_env, client):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    update = fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    assert [a.change for a in update if a.type == "SideResource"] == ["nochange", "nochange"]


@pytest.mark.asyncio
async def test_side_list(deployment, lusid_env):
    # GIVEN Side is deployed
    deployment_name = deployment.id
    scope = deployment_name
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN calling list
    result = await side_definition.SideResource.list(
        client=client,
        # No filter available
        query_params={},
        path_params={
            "scope": scope,
        }
    )
    # THEN the result has correct structure
    assert "values" in result and isinstance(result["values"], list)
    resource = result["values"][0]["value"]
    assert "id" in resource and resource["id"].startswith("side_")
    assert "side" in resource and isinstance(resource["side"], str)
    assert "scope" in resource and isinstance(resource["scope"], str)
    assert "amount" in resource and isinstance(resource["amount"], (str, dict))
    assert "currency" in resource and isinstance(resource["currency"], (str, dict))
    assert "notionalAmount" in resource and isinstance(resource["notionalAmount"], (str, dict))
    assert "rate" in resource and isinstance(resource["rate"], (str, dict))
    assert "security" in resource and isinstance(resource["security"], (str, dict))
    assert "units" in resource and isinstance(resource["units"], (str, dict))


@pytest.mark.asyncio
async def test_side_fetch(deployment, lusid_env):
    # GIVEN Side is deployed
    deployment_name = deployment.id
    scope = deployment_name
    side = "Side1"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await side_definition.SideResource.fetch(
        client=client,
        keys={
            "side": side,
            "scope": scope
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"side_{scope}_{side}".lower(),
        "side": side,
        "scope": scope,
        "amount": "Txn:TotalConsideration",
        "currency": "Txn:SettlementCurrency",
        "notionalAmount": "0",
        "rate": "Txn:TradeToPortfolioRate",
        "security": "Txn:LusidInstrumentId",
        "units": "Txn:Units"
    }
    assert response == expected_response


def test_teardown_transaction_type(deployment, lusid_env, client):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    fbnconfig.deployex(fbnconfig.Deployment(deployment.id, []), lusid_env.env, lusid_env.token)
    with pytest.raises(HTTPStatusError) as error:
        client.get(
            "/api/api/transactionconfiguration/types/default/Buy",
            params={"scope": deployment.id},
        )
        assert error.value.response.status_code == 404


def test_create_transaction_type(deployment, lusid_env, client):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    search = client.get(
        "/api/api/transactionconfiguration/types/default/Buy", params={"scope": deployment.id}
    )
    assert search.status_code == 200


def test_nochange_transaction_type(deployment, lusid_env, client):
    fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    update = fbnconfig.deployex(deployment, lusid_env.env, lusid_env.token)
    assert [a.change for a in update if a.type == "TransactionTypeResource"] == ["nochange"]


@pytest.mark.asyncio
async def test_transaction_type_list(deployment, lusid_env):
    # GIVEN TransactionType is deployed
    deployment_name = deployment.id
    scope = deployment_name
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN calling list
    result = await transaction_type.TransactionTypeResource.list(
        client=client,
        # No filter available
        query_params={},
        path_params={
            "scope": scope,
        }
    )
    # THEN the result has correct structure
    assert "values" in result and isinstance(result["values"], list)
    resource = result["values"][0]["value"]
    assert "id" in resource and resource["id"].startswith("transaction-type_")
    assert "source" in resource and isinstance(resource["source"], str)
    assert "scope" in resource and isinstance(resource["scope"], str)
    assert "aliases" in resource and isinstance(resource["aliases"], list)


@pytest.mark.asyncio
async def test_transaction_type_fetch(deployment, lusid_env):
    # GIVEN TransactionType is deployed
    deployment_name = deployment.id
    scope = deployment_name
    source = "default"
    tr_type = "Buy"
    client = fbnconfig.create_client(lusid_env.env, lusid_env.token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await transaction_type.TransactionTypeResource.fetch(
        client=client,
        keys={
            "source": source,
            "transactionType": tr_type,
            "scope": scope
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"transaction-type_{scope}_{source}_{tr_type}".lower(),
        "source": source,
        "scope": scope,
        "aliases": [
            {
                "description": "Something",
                "isDefault": False,
                "transactionClass": "default",
                "transactionRoles": "LongLonger",
                "type": "Buy",
            },
            {
                "description": "Something but different",
                "isDefault": False,
                "transactionClass": "default",
                "transactionRoles": "LongLonger",
                "type": "BY",
            }
        ],
        "calculations": [
            {
                "side": {
                    "$ref": f"side_{scope}_side2".lower()
                },
                "type": "TaxAmounts"
            },
            {
                "type": "Txn:NotionalAmount"
            }
        ],
        "movements": [
            {
                "condition": "",
                "direction": 1,
                "mappings": [],
                "movementOptions": [
                    "DirectAdjustment"
                ],
                "movementTypes": "StockMovement",
                "name": "Stock Movement",
                "properties": [],
                "settlementMode": "Internal",
                "side": {
                    "$ref": f"side_{scope}_side1".lower()
                }
            }
        ],
        "properties": []
    }
    assert response == expected_response



import copy
import json

import httpx
import pytest
from pydantic import BaseModel, Field

#
# These are not really tests, but a way of capturing how we can handle
# different client/server data behaviours using pydantic, the tests make sure it works
#
# There are a few cases:
# * a normal field that is part of the resource model on the server. Usually these are
#   required inputs and returned when you read the resource.
# * a field that we need on the resource class but isn't part of the server resource at all.
#   doesn't appear in requests or responses. The resource id (`id`) is the common case of these.
# * a field which is not controlled by the client, but calculated by the server
#   and appear in the responses. An AsAt time would be like this
# * fields the user can choose to set, but if they don't then the server calculates
#   a value for them. They might not be in the request, but will usually be in the response


class FieldTypes(BaseModel):
    # internal field that is provided by user but not sent to server
    id: str = Field(exclude=True)
    # normal field, provided by user and sent to server as-is
    data_field: str
    # something calculated on the server that the user doesn't provide (like userId or driveId)
    server_calculated: str | None = Field(None, exclude=True, init=False)
    # user can provide, but server will default if not given
    server_defaulted: str | None

    def read(self, client, old_state) -> None:
        get = client.get("/reference").json()
        get.pop("excess_field")  # field from server we don't need at all
        self._remote = get

    def create(self, client):
        desired = self.model_dump(mode="json", exclude_none=True)
        create = client.post("/reference", json=desired).json()
        self.server_calculated = create["server_calculated"]
        return {"server_calculated": self.server_calculated}

    def update(self, client, old_state):
        self.read(client, old_state)
        current = copy.deepcopy(self._remote)
        if self.server_defaulted is None:  # defaulted fields ignore if desired is None
            current.pop("server_defaulted")
        current.pop("server_calculated")  # calculated fields not compared
        desired = self.model_dump(mode="json", exclude_none=True)
        if desired == current:
            return None
        update = client.put("/reference/1", json=desired).json()
        self.server_calculated = update["server_calculated"]
        return {"server_calculated": self.server_calculated}


TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeFieldTypes:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create(self, respx_mock):
        respx_mock.post("/reference").mock(
            return_value=httpx.Response(200, json={"server_calculated": 56})
        )
        sut = FieldTypes(id="123", data_field="some data", server_defaulted=None)
        new_state = sut.create(self.client)
        assert new_state["server_calculated"] == 56
        assert sut.server_calculated == 56
        request = respx_mock.calls.last.request
        assert json.loads(request.content) == {"data_field": "some data"}

    def test_update(self, respx_mock):
        respx_mock.get("/reference").mock(
            return_value=httpx.Response(
                200,
                json={
                    "server_calculated": 45,
                    "excess_field": "not needed",
                    "data_field": "other data",
                    "server_defaulted": "server default",
                },
            )
        )
        respx_mock.put("/reference/1").mock(
            return_value=httpx.Response(200, json={"server_calculated": 45})
        )
        old_state = {}
        sut = FieldTypes(id="123", data_field="some data", server_defaulted=None)
        new_state = sut.update(self.client, old_state)
        assert new_state == {"server_calculated": 45}

    def test_update_no_change(self, respx_mock):
        respx_mock.get("/reference").mock(
            return_value=httpx.Response(
                200,
                json={
                    "server_calculated": 45,
                    "excess_field": "not needed",
                    "data_field": "some data",
                    "server_defaulted": "server default",
                },
            )
        )
        old_state = {}
        sut = FieldTypes(id="123", data_field="some data", server_defaulted=None)
        new_state = sut.update(self.client, old_state)
        assert new_state is None



from fbnconfig import Deployment, customentity


def configure(env):
    deployment_name = getattr(env, "name", "entitytype_example")
    ce = customentity.EntityTypeResource(
        id="ce2",
        entity_type_name=deployment_name,
        display_name="Takeaway menu",
        description="A menu",
        field_schema=[
            customentity.FieldDefinition(
                name="venueId",
                lifetime=customentity.LifeTime.PERPETUAL,
                type=customentity.FieldType.STRING,
                collection_type=customentity.CollectionType.SINGLE,
                required=True,
            ),
            customentity.FieldDefinition(
                name="venueOwner",
                lifetime=customentity.LifeTime.TIMEVARIANT,
                type=customentity.FieldType.STRING,
                required=True,
            ),
        ],
    )
    return Deployment(deployment_name, [ce])


if __name__ == "__main__":
    import os

    import click

    import fbnconfig

    @click.command()
    @click.argument("lusid_url", envvar="LUSID_ENV", type=str)
    @click.option("-v", "--vars_file", type=click.File("r"))
    def cli(lusid_url, vars_file):
        host_vars = fbnconfig.load_vars(vars_file)
        d = configure(host_vars)
        fbnconfig.deployex(d, lusid_url, os.environ["FBN_ACCESS_TOKEN"])

    cli()



import os
from types import SimpleNamespace

import pytest
from pytest import fixture

import fbnconfig
from fbnconfig import Deployment, horizon, property
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("horizon")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    print("\nTearing down resources...")
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


@fixture(scope="module")
def coppclark_instance(setup_deployment):
    return horizon.IntegrationInstanceResource(
        id="test-horizon-instance",
        integration_type="copp-clark",
        name=f"test-integration-{setup_deployment.name}",
        description="Test integration instance for horizon tests",
        enabled=True,
        triggers=[
            horizon.Trigger(type="time", cron_expression="0 0 9 ? * MON-FRI", time_zone="Europe/London")
        ],
        details={
            "paymentSystemsCalendar": {"currencyFilter": ["GBP", "USD"], "importUnqualified": False}
        },
    )


@fixture(scope="module")
def aladdin_instance(setup_deployment):
    return horizon.IntegrationInstanceResource(
        id="aladdin-instance",
        integration_type="blackrock-aladdin",
        name=f"test_aladdin-{setup_deployment.name}",
        description="my first aladdin",
        enabled=False,
        triggers=[],
        details={
            "sourceFileLocation": "/sftp",
            "onboardingDate": "2022-01-01",
            "localTimeZone": "Australia/Melbourne",
            "unpackFromArchive": False,
            "archiveFileMask": "",
            "isFxNdfAllowed": True,
            "int29AnalyticsFileType": "NotApplicable",
            "allowCoreFieldUpdates": "Never",
            "blackRockAladdinInterfaceSelections": [{
                "interfaceNumber": "Int96",
                "fileMask": "^ATOM_R_newcash96_daily.*_*.xml$",
                "localCutTime": "00:00"
            }]

        }
    )


@fixture
def updated_instance(setup_deployment):
    return horizon.IntegrationInstanceResource(
        id="test-update-instance",
        integration_type="copp-clark",
        name=f"updated-integration-{setup_deployment.name}",
        description="Updated test integration instance",
        enabled=True,
        triggers=[horizon.Trigger(type="time", cron_expression="0 0 12 ? * *", time_zone="UTC")],
        details={
            "paymentSystemsCalendar": {"currencyFilter": ["EUR", "JPY"], "importUnqualified": True}
        },
    )


@fixture(scope="module")
def property_ref(setup_deployment):
    # Instrument/default/Status exists in every domain
    return property.DefinitionRef(
        id="test-prop-ref", domain=property.Domain.Instrument, scope="default", code="Status"
    )


@fixture(scope="module")
def optional_props(coppclark_instance, property_ref):
    return horizon.OptionalPropsResource(
        id="test-horizon-props",
        instance=coppclark_instance,
        props=[
            horizon.OptionalProp(
                property=property_ref,
                display_name_override="Test Status Override",
                description_override="Test description override",
                entity_type="Instrument",
                entity_sub_type=["Bond", "Equity"],
                vendor_package=["TestVendor"],
            )
        ],
    )


@fixture(scope="module")
def full_deployment(setup_deployment, coppclark_instance, optional_props, aladdin_instance):
    """Create a full deployment with instance and props"""
    return Deployment(setup_deployment.name, [coppclark_instance, optional_props, aladdin_instance])


@fixture
def update_deployment(setup_deployment, updated_instance):
    """Create an update deployment"""
    return Deployment(setup_deployment.name, [updated_instance])


def test_create_integration_instance(setup_deployment, full_deployment):
    """Test creating a horizon integration instance"""
    fbnconfig.deploy(full_deployment, lusid_env, token)
    instances = client.get("/horizon/api/integrations/instances").json()
    instance_names = [instance["name"] for instance in instances]
    expected_name = f"test-integration-{setup_deployment.name}"
    assert expected_name in instance_names


def test_update_integration_instance(setup_deployment, full_deployment, update_deployment):
    """Test updating an integration instance"""
    fbnconfig.deploy(full_deployment, lusid_env, token)
    update = fbnconfig.deploy(update_deployment, lusid_env, token)
    instances = client.get("/horizon/api/integrations/instances").json()
    instance_names = [instance["name"] for instance in instances]
    expected_name = f"updated-integration-{setup_deployment.name}"
    assert expected_name in instance_names
    # Check that the instance was updated and props were removed
    instance_changes = [a.change for a in update if a.type == "IntegrationInstanceResource"]
    props_changes = [a.change for a in update if a.type == "OptionalPropsResource"]
    assert "update" in instance_changes or "create" in instance_changes
    assert "remove" in props_changes


def test_teardown_integration_instance(setup_deployment, full_deployment):
    """Test removing integration instances"""
    fbnconfig.deploy(full_deployment, lusid_env, token)
    instances_before = client.get("/horizon/api/integrations/instances").json()
    initial_count = len(instances_before)
    teardown = fbnconfig.deploy(fbnconfig.Deployment(setup_deployment.name, []), lusid_env, token)
    instances_after = client.get("/horizon/api/integrations/instances").json()
    final_count = len(instances_after)
    assert final_count < initial_count
    # Check that resources were removed
    instance_changes = [a.change for a in teardown if a.type == "IntegrationInstanceResource"]
    props_changes = [a.change for a in teardown if a.type == "OptionalPropsResource"]
    assert instance_changes == ["remove"] * 2
    assert props_changes == ["remove"]


def test_optional_props_configuration(setup_deployment, full_deployment):
    """Test configuring optional properties for horizon integration"""
    fbnconfig.deploy(full_deployment, lusid_env, token)
    instances = client.get("/horizon/api/integrations/instances").json()
    test_instance = next(
        (i for i in instances if i["name"] == f"test-integration-{setup_deployment.name}"), None
    )
    assert test_instance is not None
    instance_id = test_instance["id"]
    config_url = f"/horizon/api/integrations/instances/configuration/copp-clark/{instance_id}"
    config = client.get(config_url).json()
    assert "Instrument/TestStatus" in config or len(config) >= 0


def test_integration_instance_with_triggers(setup_deployment, full_deployment):
    """Test that triggers are properly configured"""
    fbnconfig.deploy(full_deployment, lusid_env, token)
    instances = client.get("/horizon/api/integrations/instances").json()
    test_instance = next(
        (i for i in instances if i["name"] == f"test-integration-{setup_deployment.name}"), None
    )
    assert test_instance is not None
    assert "triggers" in test_instance
    assert len(test_instance["triggers"]) > 0
    trigger = test_instance["triggers"][0]
    assert trigger["type"] == "time"
    assert trigger["cronExpression"] == "0 0 9 ? * MON-FRI"
    assert trigger["timeZone"] == "Europe/London"


def test_integration_instance_details(setup_deployment, full_deployment):
    """Test that instance details are properly configured"""
    fbnconfig.deploy(full_deployment, lusid_env, token)
    instances = client.get("/horizon/api/integrations/instances").json()
    test_instance = next(
        (i for i in instances if i["name"] == f"test-integration-{setup_deployment.name}"), None
    )
    assert test_instance is not None
    assert "details" in test_instance
    details = test_instance["details"]
    assert "paymentSystemsCalendar" in details
    calendar_config = details["paymentSystemsCalendar"]
    assert "currencyFilter" in calendar_config
    assert "GBP" in calendar_config["currencyFilter"]
    assert "USD" in calendar_config["currencyFilter"]
    assert calendar_config["importUnqualified"] is False


def test_update_nochanges(setup_deployment, full_deployment):
    """Test that deploying the same configuration twice doesn't cause issues"""
    fbnconfig.deploy(full_deployment, lusid_env, token)
    instances_first = client.get("/horizon/api/integrations/instances").json()
    first_count = len(instances_first)
    update = fbnconfig.deploy(full_deployment, lusid_env, token)
    instances_second = client.get("/horizon/api/integrations/instances").json()
    second_count = len(instances_second)
    assert first_count == second_count
    # Check that no changes were made on second deployment
    instance_changes = [a.change for a in update if a.type == "IntegrationInstanceResource"]
    props_changes = [a.change for a in update if a.type == "OptionalPropsResource"]
    assert instance_changes == ["nochange"] * 2
    assert props_changes == ["nochange"]


@pytest.mark.asyncio
async def test_integration_instance_list(setup_deployment):
    # GIVEN IntegrationInstance resource is deployed
    deployment_name = setup_deployment.name
    # WHEN calling list with filter for the deployment name
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    result = await horizon.IntegrationInstanceResource.list(
        client=client,
        query_params={}
    )
    instance_name = f"test-integration-{deployment_name}"
    instance = next(
        (i for i in result["values"] if i["value"]["name"] == instance_name)
    )
    instance_id = instance["keys"]["instanceId"]
    # THEN the result is as expected
    expected_instance = {
        "keys": {
            "instanceId": instance_id,
        },
        "value": {
            "id": f"integration-instance_{instance_id}".lower(),
            "instanceId": instance_id,
            "name": instance_name,
            "integrationType": "copp-clark",
            "description": "Test integration instance for horizon tests",
            "enabled": True,
            "details": {
                "paymentSystemsCalendar": {
                    "currencyFilter": ["GBP", "USD"],
                    "importUnqualified": False
                }
            },
            "triggers": [
                {
                    "type": "time",
                    "cronExpression": "0 0 9 ? * MON-FRI",
                    "timeZone": "Europe/London"
                }
            ]
        }
    }
    assert instance == expected_instance


@pytest.mark.asyncio
async def test_integration_instance_fetch(setup_deployment):
    # GIVEN IntegrationInstance resource is deployed
    deployment_name = setup_deployment.name
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    instances_response = await client.get("/horizon/api/integrations/instances")
    instances = instances_response.json()
    instance_name = f"test-integration-{deployment_name}"
    instance = next(
        (i for i in instances if i["name"] == instance_name)
    )
    instance_id = instance["id"]
    # WHEN the fetch method is called for that resource
    response = await horizon.IntegrationInstanceResource.fetch(
        client=client,
        keys={
            "instanceId": instance_id,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"integration-instance_{instance_id}".lower(),
        "instanceId": instance_id,
        "name": instance_name,
        "integrationType": "copp-clark",
        "description": "Test integration instance for horizon tests",
        "enabled": True,
        "details": {
            "paymentSystemsCalendar": {
                "currencyFilter": ["GBP", "USD"],
                "importUnqualified": False
            }
        },
        "triggers": [
            {
                "type": "time",
                "cronExpression": "0 0 9 ? * MON-FRI",
                "timeZone": "Europe/London"
            }
        ]
    }
    assert response == expected_response


@pytest.mark.asyncio
async def test_optional_props_fetch(setup_deployment):
    # GIVEN OptionalProps resource is deployed
    deployment_name = setup_deployment.name
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    instances_response = await client.get("/horizon/api/integrations/instances")
    instances = instances_response.json()
    instance_name = f"test-integration-{deployment_name}"
    instance = next(
        (i for i in instances if i["name"] == instance_name)
    )
    instance_id = instance["id"]
    integration_type = "copp-clark"
    # WHEN the fetch method is called for that resource
    response = await horizon.OptionalPropsResource.fetch(
        client=client,
        keys={
            "instanceId": instance_id,
            "instanceIntegrationType": integration_type
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"integration-instance-optional-props_{integration_type}_{instance_id}".lower(),
        "instance": {
            "$ref": f"integration-instance_{instance_id}".lower()
        },
        "props": []
    }
    assert response == expected_response



from fbnconfig import Deployment, drive, scheduler


def configure(env):
    deployment_name = getattr(env, "name", "mixed")

    client_folder = drive.FolderResource(id="base_folder", name=env.base_dir, parent=drive.root)
    files_folder = drive.FolderResource(id="files", name="files", parent=client_folder)
    content = """
    a little poem
        about ducks
            quack quack
    """
    data_file = drive.FileResource(id="file1", folder=files_folder, name="mydata.txt", content=content)
    version = 101
    docker_image = scheduler.ImageResource(
        id="img_mixed",
        source_image="harbor.finbourne.com/ceng/fbnconfig-pipeline:0.1",
        dest_name="beany_mixed",
        dest_tag=f"v{version}",
    )
    data_job = scheduler.JobResource(
        id="data_job",
        scope=deployment_name,
        code="mixed-job",
        image=docker_image,
        name="myjob",
        description="banana",
    )
    data_sched = scheduler.ScheduleResource(
        id="data_sch",
        name="my-schedule",
        scope=deployment_name,
        code="data-sched",
        job=data_job,
        expression="0 50 4 * * ? *",
        timezone="Europe/London",
        description="whatever",
    )
    return Deployment(
        deployment_name, [client_folder, files_folder, data_file, docker_image, data_job, data_sched]
    )



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import datatype
from fbnconfig import relational_dataset_definition as rdd

TEST_BASE = "https://foo.lusid.com"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeRelationalDatasetDefinitionRef:
    base_url = TEST_BASE
    dataset_def_url = "/api/api/relationaldatasetdefinitions"
    client = httpx.Client(
        base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]}
    )

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get(f"{self.dataset_def_url}/default/sample-dataset").mock(
            side_effect=[
                httpx.Response(
                    200, json={"id": {"scope": "default", "code": "sample-dataset"}}
                )
            ]
        )

        sut = rdd.RelationalDatasetDefinitionRef(
            id="one", scope="default", code="sample-dataset"
        )
        # when we call attach
        sut.attach(self.client)
        # then a get request was made
        req = respx_mock.calls.last.request
        assert req.url == f"{self.base_url + self.dataset_def_url}/default/sample-dataset"

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get(f"{self.dataset_def_url}/default/missing-dataset").mock(
            side_effect=[
                httpx.Response(
                    404, json={"name": "RelationalDatasetDefinitionNotFound"}
                )
            ]
        )
        sut = rdd.RelationalDatasetDefinitionRef(
            id="two", scope="default", code="missing-dataset"
        )

        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(self.client)

        assert (
            "Relational dataset definition default/missing-dataset does not exist"
            in str(ex.value)
        )

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get(f"{self.dataset_def_url}/default/error-dataset").mock(
            side_effect=[httpx.Response(400, json={})]
        )
        sut = rdd.RelationalDatasetDefinitionRef(
            id="three", scope="default", code="error-dataset"
        )
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(self.client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeRelationalDatasetDefinitionResource:
    base_url = TEST_BASE
    dataset_def_url = "/api/api/relationaldatasetdefinitions"
    client = httpx.Client(
        base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]}
    )

    def test_create_dataset_definition(self, respx_mock):
        # GIVEN Resource does not exist
        respx_mock.post(self.dataset_def_url).mock(
            side_effect=[
                httpx.Response(
                    status_code=201,
                    json={
                        "id": {"scope": "default", "code": "sample-dataset"},
                        "displayName": "Sample Financial Dataset",
                        "description": "A sample relational dataset definition",
                        "applicableEntityTypes": ["Instrument", "Portfolio"],
                        "fieldSchema": [
                            {
                                "fieldName": "instrumentId",
                                "displayName": "Instrument ID",
                                "description": "The unique identifier",
                                "dataTypeId": {"scope": "system", "code": "string"},
                                "required": True,
                                "category": "SeriesIdentifier",
                            }
                        ],
                        "href": "https://example.com/api/relationaldatasetdefinitions/default/sample-dataset",
                        "version": {
                            "effectiveFrom": "2019-01-01T12:00:00.0000000+00:00",
                            "asAtDate": "2019-01-01T12:00:00.0100000+00:00",
                        },
                    },
                )
            ]
        )

        # WHEN Dataset Definition Resource is created
        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="sample-dataset",
            scope="default",
            code="sample-dataset",
            display_name="Sample Financial Dataset",
            description="A sample relational dataset definition",
            applicable_entity_types=[
                rdd.ApplicableEntityType.INSTRUMENT,
                rdd.ApplicableEntityType.PORTFOLIO
            ],
            field_schema=[
                rdd.FieldSchema(
                    field_name="instrumentId",
                    display_name="Instrument ID",
                    description="The unique identifier",
                    data_type_id=datatype.ResourceId(scope="system", code="string"),
                    required=True,
                    category=rdd.FieldCategory.SERIES_IDENTIFIER,
                )
            ],
        )

        new_state = dataset_def.create(self.client)

        # THEN State returned as expected
        assert new_state == {
            "scope": "default",
            "code": "sample-dataset",
        }

        # and the expected post request was sent
        request: httpx.Request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == self.dataset_def_url
        assert f"https://{request.url.host}" == self.base_url
        assert json.loads(request.content) == {
            "id": {"scope": "default", "code": "sample-dataset"},
            "displayName": "Sample Financial Dataset",
            "description": "A sample relational dataset definition",
            "applicableEntityTypes": ["Instrument", "Portfolio"],
            "fieldSchema": [
                {
                    "fieldName": "instrumentId",
                    "displayName": "Instrument ID",
                    "description": "The unique identifier",
                    "dataTypeId": {"scope": "system", "code": "string"},
                    "required": True,
                    "category": "SeriesIdentifier",
                }
            ],
        }

    def test_create_dataset_definition_with_datatypekey(self, respx_mock):

        system_string_datatype = datatype.DataTypeRef(id="systemstring", scope="system", code="string")

        # GIVEN Resource does not exist
        respx_mock.post(self.dataset_def_url).mock(
            side_effect=[
                httpx.Response(
                    status_code=201,
                    json={
                        "id": {"scope": "default", "code": "sample-dataset"},
                        "displayName": "Sample Financial Dataset",
                        "description": "A sample relational dataset definition",
                        "applicableEntityTypes": ["Instrument", "Portfolio"],
                        "fieldSchema": [
                            {
                                "fieldName": "instrumentId",
                                "displayName": "Instrument ID",
                                "description": "The unique identifier",
                                "dataTypeId": {"scope": "system", "code": "string"},
                                "required": True,
                                "category": "SeriesIdentifier",
                            }
                        ],
                        "href": "https://example.com/api/relationaldatasetdefinitions/default/sample-dataset",
                        "version": {
                            "effectiveFrom": "2019-01-01T12:00:00.0000000+00:00",
                            "asAtDate": "2019-01-01T12:00:00.0100000+00:00",
                        },
                    },
                )
            ]
        )

        # WHEN Dataset Definition Resource is created
        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="sample-dataset",
            scope="default",
            code="sample-dataset",
            display_name="Sample Financial Dataset",
            description="A sample relational dataset definition",
            applicable_entity_types=[
                rdd.ApplicableEntityType.INSTRUMENT,
                rdd.ApplicableEntityType.PORTFOLIO
            ],
            field_schema=[
                rdd.FieldSchema(
                    field_name="instrumentId",
                    display_name="Instrument ID",
                    description="The unique identifier",
                    data_type_id=system_string_datatype,
                    required=True,
                    category=rdd.FieldCategory.SERIES_IDENTIFIER,
                )
            ],
        )

        new_state = dataset_def.create(self.client)

        # THEN State returned as expected
        assert new_state == {
            "scope": "default",
            "code": "sample-dataset",
        }

        # and the expected post request was sent
        request: httpx.Request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == self.dataset_def_url
        assert f"https://{request.url.host}" == self.base_url
        assert json.loads(request.content) == {
            "id": {"scope": "default", "code": "sample-dataset"},
            "displayName": "Sample Financial Dataset",
            "description": "A sample relational dataset definition",
            "applicableEntityTypes": ["Instrument", "Portfolio"],
            "fieldSchema": [
                {
                    "fieldName": "instrumentId",
                    "displayName": "Instrument ID",
                    "description": "The unique identifier",
                    "dataTypeId": {"scope": "system", "code": "string"},
                    "required": True,
                    "category": "SeriesIdentifier",
                }
            ],
        }

    def test_update_no_change(self, respx_mock):

        # WHEN Dataset Definition Resource is created
        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="sample-dataset",
            scope="default",
            code="sample-dataset-no-change",
            display_name="Sample Financial Dataset",
            description="A sample relational dataset definition",
            applicable_entity_types=[
                rdd.ApplicableEntityType.INSTRUMENT,
                rdd.ApplicableEntityType.PORTFOLIO
            ],
            field_schema=[
                rdd.FieldSchema(
                    field_name="instrumentId",
                    display_name="Instrument ID",
                    description="The unique identifier",
                    data_type_id=datatype.ResourceId(scope="system", code="string"),
                    required=True,
                    category=rdd.FieldCategory.SERIES_IDENTIFIER,
                )
            ],
        )

        respx_mock.get(
            f"{self.dataset_def_url}/{dataset_def.scope}/{dataset_def.code}"
        ).mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "id": {"scope": "default", "code": "sample-dataset-no-change"},
                        "displayName": dataset_def.display_name,
                        "description": dataset_def.description,
                        "applicableEntityTypes": dataset_def.applicable_entity_types,
                        "fieldSchema": [
                            {
                                "fieldName": field.field_name,
                                "displayName": field.display_name,
                                "description": field.description,
                                "dataTypeId": {
                                    "scope": field.data_type_id.scope,
                                    "code": field.data_type_id.code
                                },
                                "required": field.required,
                                "category": field.category,
                            }
                            for field in dataset_def.field_schema
                        ],
                        "version": {
                            "effectiveFrom": "2019-01-01T12:00:00.0000000+00:00",
                            "asAtDate": "2019-01-01T12:00:00.0100000+00:00",
                        },
                        "href": "https://example.com/api/relationaldatasetdefinitions/default/sample-dataset-no-change",
                    },
                )
            ]
        )

        old_state = SimpleNamespace(
            id=dataset_def.id, scope=dataset_def.scope, code=dataset_def.code
        )

        new_state = dataset_def.update(self.client, old_state)
        # THEN no update is made
        assert new_state is None

    def test_update_no_change_null_description(self, respx_mock):

        # WHEN Dataset Definition Resource is created
        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="sample-dataset",
            scope="default",
            code="sample-dataset-no-change",
            display_name="Sample Financial Dataset",
            description="A sample relational dataset definition",
            applicable_entity_types=[
                rdd.ApplicableEntityType.INSTRUMENT,
                rdd.ApplicableEntityType.PORTFOLIO
            ],
            field_schema=[
                rdd.FieldSchema(
                    field_name="instrumentId",
                    display_name="Instrument ID",
                    description="The unique identifier",
                    data_type_id=datatype.ResourceId(scope="system", code="string"),
                    required=True,
                    category=rdd.FieldCategory.SERIES_IDENTIFIER,
                )
            ],
        )

        # AND we set the description to None
        dataset_def.description = None

        respx_mock.get(
            f"{self.dataset_def_url}/{dataset_def.scope}/{dataset_def.code}"
        ).mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "id": {"scope": "default", "code": "sample-dataset-no-change"},
                        "displayName": dataset_def.display_name,
                        "description": dataset_def.description,
                        "applicableEntityTypes": dataset_def.applicable_entity_types,
                        "fieldSchema": [
                            {
                                "fieldName": field.field_name,
                                "displayName": field.display_name,
                                "description": field.description,
                                "dataTypeId": {
                                    "scope": field.data_type_id.scope,
                                    "code": field.data_type_id.code
                                },
                                "required": field.required,
                                "category": field.category,
                            }
                            for field in dataset_def.field_schema
                        ],
                        "version": {
                            "effectiveFrom": "2019-01-01T12:00:00.0000000+00:00",
                            "asAtDate": "2019-01-01T12:00:00.0100000+00:00",
                        },
                        "href": "https://example.com/api/relationaldatasetdefinitions/default/sample-dataset-no-change",
                    },
                )
            ]
        )

        old_state = SimpleNamespace(
            id=dataset_def.id, scope=dataset_def.scope, code=dataset_def.code
        )

        new_state = dataset_def.update(self.client, old_state)
        # THEN no update is made
        assert new_state is None

    def test_update_identifier(self, respx_mock):
        # GIVEN Dataset definition exists
        # WHEN updating scope or code
        # THEN Dataset is deleted and recreated
        old_state = SimpleNamespace(id="dd1", scope="default", code="old-dataset")

        respx_mock.delete(
            f"{self.dataset_def_url}/{old_state.scope}/{old_state.code}"
        ).mock(side_effect=[httpx.Response(200, json={"asAt": "asAt deleted"})])

        respx_mock.post(self.dataset_def_url).mock(
            side_effect=[
                httpx.Response(
                    status_code=201,
                    json={
                        "id": {"scope": "new-scope", "code": "new-dataset"},
                        "version": {"asAtDate": "2019-01-01T12:00:00.0100000+00:00"},
                    },
                )
            ]
        )

        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="dd1",
            scope="new-scope",
            code="new-dataset",
            display_name="Sample Financial Dataset",
            description="A sample relational dataset definition",
            applicable_entity_types=[
                rdd.ApplicableEntityType.INSTRUMENT,
                rdd.ApplicableEntityType.PORTFOLIO
            ],
            field_schema=[
                rdd.FieldSchema(
                    field_name="instrumentId",
                    display_name="Instrument ID",
                    description="The unique identifier",
                    data_type_id=datatype.ResourceId(scope="system", code="string"),
                    required=True,
                    category=rdd.FieldCategory.SERIES_IDENTIFIER,
                )
            ],
        )

        new_state = dataset_def.update(self.client, old_state)
        assert new_state == {
            "scope": "new-scope",
            "code": "new-dataset",
        }

        # and the post is made to create the dataset
        request = respx_mock.calls.last.request
        assert request.method == "POST"

    def test_update_display_name(self, respx_mock):
        # GIVEN Dataset definition exists
        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="dd1",
            scope="default",
            code="sample-dataset",
            display_name="Original Name",
            description="A sample relational dataset definition",
            applicable_entity_types=[
                rdd.ApplicableEntityType.INSTRUMENT,
                rdd.ApplicableEntityType.PORTFOLIO
            ],
            field_schema=[
                rdd.FieldSchema(
                    field_name="instrumentId",
                    display_name="Instrument ID",
                    description="The unique identifier",
                    data_type_id=datatype.ResourceId(scope="system", code="string"),
                    required=True,
                    category=rdd.FieldCategory.SERIES_IDENTIFIER,
                )
            ],
        )

        respx_mock.get(
            f"{self.dataset_def_url}/{dataset_def.scope}/{dataset_def.code}"
        ).mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "id": {"scope": "default", "code": "sample-dataset"},
                        "displayName": "Original Name",
                        "description": dataset_def.description,
                        "applicableEntityTypes": dataset_def.applicable_entity_types,
                        "fieldSchema": [
                            {
                                "fieldName": field.field_name,
                                "displayName": field.display_name,
                                "description": field.description,
                                "dataTypeId": {"scope": "system", "code": "string"},
                                "required": field.required,
                                "category": field.category,
                            }
                            for field in dataset_def.field_schema
                        ],
                        "version": {"asAtDate": "2019-01-01T12:00:00.0100000+00:00"},
                        "href": "https://example.com",
                    },
                )
            ]
        )

        respx_mock.put(
            f"{self.dataset_def_url}/{dataset_def.scope}/{dataset_def.code}"
        ).mock(
            side_effect=[
                httpx.Response(
                    status_code=200,
                    json={
                        "displayName": "Updated Name",
                        "version": {"asAtDate": "2019-01-01T13:00:00.0100000+00:00"},
                    },
                )
            ]
        )

        old_state = SimpleNamespace(
            id="dd1", scope="default", code="sample-dataset"
        )

        # WHEN display name changes
        dataset_def.display_name = "Updated Name"
        new_state = dataset_def.update(self.client, old_state)

        # THEN update is made
        assert new_state == {
            "scope": "default",
            "code": "sample-dataset",
        }

        request = respx_mock.calls.last.request
        assert request.method == "PUT"

    def test_delete(self, respx_mock):
        respx_mock.delete(f"{self.dataset_def_url}/default/sample-dataset").mock(
            side_effect=[httpx.Response(200, json={"asAt": "asAt deleted"})]
        )

        old_state = SimpleNamespace(
            id="dd1", scope="default", code="sample-dataset"
        )

        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="dd1",
            scope="default",
            code="sample-dataset",
            display_name="Original Name",
            description="A sample relational dataset definition",
            applicable_entity_types=[
                rdd.ApplicableEntityType.INSTRUMENT,
                rdd.ApplicableEntityType.PORTFOLIO
            ],
            field_schema=[
                rdd.FieldSchema(
                    field_name="instrumentId",
                    display_name="Instrument ID",
                    description="The unique identifier",
                    data_type_id=datatype.ResourceId(scope="system", code="string"),
                    required=True,
                    category=rdd.FieldCategory.SERIES_IDENTIFIER,
                )
            ],
        )

        assert dataset_def.delete(self.client, old_state) is None

        request = respx_mock.calls.last.request
        assert request.method == "DELETE"

    @staticmethod
    def test_deps_with_datatyperef():
        # given a dataset definition with a system/string dataTypeId
        system_string_datatype = datatype.DataTypeRef(id="systemstring", scope="system", code="string")
        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="dd1",
            scope="default",
            code="dataset-with-system-types",
            display_name="Dataset with System Types",
            applicable_entity_types=[rdd.ApplicableEntityType.INSTRUMENT],
            field_schema=[
                rdd.FieldSchema(
                    field_name="stringField",
                    display_name="String Field",
                    data_type_id=system_string_datatype,
                    required=True,
                    category=rdd.FieldCategory.VALUE,
                )
            ],
        )

        # when we get dependencies
        deps = dataset_def.deps()

        # then we get back the system string data type ID
        assert deps == [system_string_datatype]

    @staticmethod
    def test_deps_with_datatype_resource():
        # given a dataset definition with a custom DataTypeResource
        custom_type = datatype.DataTypeResource(
            id="dt",
            scope="dtscope",
            code="dtcode",
            type_value_range=datatype.TypeValueRange.CLOSED,
            display_name="Priority Test",
            description="A test datatype for Priority",
            value_type=datatype.ValueType.STRING,
            acceptable_values=["High", "Medium", "Low"],
        )

        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="dd1",
            scope="default",
            code="dataset-with-custom-type",
            display_name="Dataset with Custom Type",
            applicable_entity_types=[rdd.ApplicableEntityType.INSTRUMENT],
            field_schema=[
                rdd.FieldSchema(
                    field_name="customField",
                    display_name="Custom Field",
                    data_type_id=custom_type,
                    required=True,
                    category=rdd.FieldCategory.VALUE,
                )
            ],
        )

        # when we get dependencies
        deps = dataset_def.deps()

        # then it includes the custom data type
        assert deps == [custom_type]

    @staticmethod
    def test_deps_with_resource_id():
        # given a dataset definition with ResourceId (system types)
        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="dd1",
            scope="default",
            code="dataset-with-system-types",
            display_name="Dataset with System Types",
            applicable_entity_types=[rdd.ApplicableEntityType.INSTRUMENT],
            field_schema=[
                rdd.FieldSchema(
                    field_name="stringField",
                    display_name="String Field",
                    data_type_id=datatype.ResourceId(scope="system", code="string"),
                    required=True,
                    category=rdd.FieldCategory.VALUE,
                )
            ],
        )

        # when we get dependencies
        deps = dataset_def.deps()

        # then no dependencies (ResourceId is not a Resource or Ref)
        assert deps == []

    @staticmethod
    def test_deps_with_multiple_fields():
        # given a dataset definition with multiple fields and mixed data types
        custom_type = datatype.DataTypeResource(
            id="dt",
            scope="dtscope",
            code="dtcode",
            type_value_range=datatype.TypeValueRange.CLOSED,
            display_name="Priority Test",
            description="A test datatype for Priority",
            value_type=datatype.ValueType.STRING,
            acceptable_values=["High", "Medium", "Low"],
        )

        type_ref = datatype.DataTypeRef(
            id="type-ref", scope="system", code="decimal"
        )

        dataset_def = rdd.RelationalDatasetDefinitionResource(
            id="dd1",
            scope="default",
            code="multi-field-dataset",
            display_name="Multi-field Dataset",
            applicable_entity_types=[
                rdd.ApplicableEntityType.INSTRUMENT,
                rdd.ApplicableEntityType.PORTFOLIO
            ],
            field_schema=[
                rdd.FieldSchema(
                    field_name="field1",
                    display_name="Field 1",
                    data_type_id=custom_type,
                    required=True,
                    category=rdd.FieldCategory.VALUE,
                ),
                rdd.FieldSchema(
                    field_name="field2",
                    display_name="Field 2",
                    data_type_id=datatype.ResourceId(scope="system", code="string"),
                    required=False,
                    category=rdd.FieldCategory.SERIES_IDENTIFIER,
                ),
                rdd.FieldSchema(
                    field_name="field3",
                    display_name="Field 3",
                    data_type_id=type_ref,
                    required=True,
                    category=rdd.FieldCategory.METADATA,
                ),
            ],
        )

        # when we get dependencies
        deps = dataset_def.deps()

        # then it includes both the resource and ref, but not the ResourceId
        assert deps == [custom_type, type_ref]

    def test_dump_simple_dataset_definition(self):
        # given a simple dataset definition resource
        sut = rdd.RelationalDatasetDefinitionResource(
            id="dump-dataset",
            scope="test-scope",
            code="DumpDataset",
            display_name="Test Dataset",
            description="A test dataset for dumping",
            applicable_entity_types=[rdd.ApplicableEntityType.INSTRUMENT],
            field_schema=[
                rdd.FieldSchema(
                    field_name="price",
                    display_name="Price",
                    data_type_id=datatype.ResourceId(scope="system", code="number"),
                    required=True,
                    category=rdd.FieldCategory.VALUE,
                )
            ],
        )
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"},
        )
        # then all fields are included (no excludes)
        expected = {
            "scope": "test-scope",
            "code": "DumpDataset",
            "displayName": "Test Dataset",
            "description": "A test dataset for dumping",
            "applicableEntityTypes": ["Instrument"],
            "fieldSchema": [
                {
                    "fieldName": "price",
                    "displayName": "Price",
                    "dataTypeId": {"scope": "system", "code": "number"},
                    "required": True,
                    "category": "Value",
                }
            ],
        }
        assert result == expected

    def test_undump_simple_dataset_definition(self):
        # given dumped dataset definition data with id context
        data = {
            "scope": "test-scope",
            "code": "DumpDataset",
            "displayName": "Test Dataset",
            "description": "A test dataset for undumping",
            "applicableEntityTypes": ["Portfolio"],
            "fieldSchema": [
                {
                    "fieldName": "nav",
                    "displayName": "NAV",
                    "dataTypeId": {"scope": "system", "code": "number"},
                    "required": False,
                    "category": "Value",
                }
            ],
        }
        # when we undump it with id context
        result = rdd.RelationalDatasetDefinitionResource.model_validate(
            data, context={"id": "undump-dataset"}
        )
        # then the resource is properly constructed
        assert result.id == "undump-dataset"
        assert result.scope == "test-scope"
        assert result.code == "DumpDataset"
        assert result.display_name == "Test Dataset"
        assert result.description == "A test dataset for undumping"
        assert result.applicable_entity_types == ["Portfolio"]
        assert len(result.field_schema) == 1
        assert result.field_schema[0].field_name == "nav"
        assert result.field_schema[0].required is False

    def test_from_api(self, testdata_loader):
        input_data = testdata_loader(
            "relational_dataset_definitions",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "relational_dataset_definitions",
            "list_expected_result"
        )["values"][0]["value"]
        result_data = rdd.RelationalDatasetDefinitionResource.from_api(input_data)
        assert result_data == expected_result

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        input_data = testdata_loader(
            "relational_dataset_definitions",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "relational_dataset_definitions",
            "list_expected_result"
        )
        respx_mock.get(rdd.RelationalDatasetDefinitionResource.list_endpoint()).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await rdd.RelationalDatasetDefinitionResource.list(client, query_params={})
            assert result == expected_result

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        input_data = testdata_loader(
            "relational_dataset_definitions",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "relational_dataset_definitions",
            "list_expected_result"
        )["values"][0]["value"]
        keys = {"scope": "scope", "code": "code"}
        respx_mock.get(rdd.RelationalDatasetDefinitionResource.fetch_endpoint().format(**keys)).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await rdd.RelationalDatasetDefinitionResource.fetch(client, keys=keys)
        assert result == expected_result


class DescribeFieldSchema:
    @staticmethod
    def test_field_schema_serialization():
        # given a field schema
        field = rdd.FieldSchema(
            field_name="instrumentId",
            display_name="Instrument ID",
            description="The unique identifier",
            data_type_id=datatype.DataTypeRef(id="dt", scope="system", code="string"),
            required=True,
            category=rdd.FieldCategory.SERIES_IDENTIFIER,
        )

        # when we serialize it
        result = field.model_dump(mode="json", by_alias=True, exclude_none=True)

        # then it matches the expected format
        assert result == {
            "fieldName": "instrumentId",
            "displayName": "Instrument ID",
            "description": "The unique identifier",
            "dataTypeId": {"scope": "system", "code": "string"},
            "required": True,
            "category": "SeriesIdentifier",
        }

    @staticmethod
    def test_field_schema_with_optional_fields():
        # given a minimal field schema
        field = rdd.FieldSchema(
            field_name="optionalField",
            display_name="Optional Field",
            data_type_id=datatype.ResourceId(scope="system", code="string"),
            category=rdd.FieldCategory.METADATA,
        )

        # when we serialize it
        result = field.model_dump(mode="json", by_alias=True, exclude_none=True)

        # then optional fields are excluded and required is automatically set to True
        assert result == {
            "fieldName": "optionalField",
            "displayName": "Optional Field",
            "dataTypeId": {"scope": "system", "code": "string"},
            "required": True,
            "category": "Metadata",
        }



import asyncio
import time
from typing import Literal, overload

import httpx

# Throw an exception whenever the response code is 4xx or 5xx


def response_hook(response: httpx.Response) -> None:
    if response.is_error:
        response.read()
        response.raise_for_status()


async def response_hook_async(response: httpx.Response) -> None:
    if response.is_error:
        await response.aread()
        response.raise_for_status()


class RetryTransport(httpx.BaseTransport):
    def __init__(self, **kwargs):
        self._wrapper = httpx.HTTPTransport(**kwargs)

    def handle_request(self, request):
        response = self._wrapper.handle_request(request)
        if response.status_code == 429:
            retry_header = response.headers.get("retry-after", None)
            delay = int(retry_header) if retry_header else 0.5
            time.sleep(delay)
            return self.handle_request(request)
        return response

    def close(self):
        self._wrapper.close()


class AsyncRetryTransport(httpx.AsyncBaseTransport):
    def __init__(self, **kwargs):
        self._wrapper = httpx.AsyncHTTPTransport(**kwargs)

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        response = await self._wrapper.handle_async_request(request)
        if response.status_code == 429:
            retry_header = response.headers.get("retry-after")
            delay = float(retry_header) if retry_header else 0.5
            await response.aread()
            await asyncio.sleep(delay)
            return await self.handle_async_request(request)
        return response

    async def aclose(self) -> None:
        await self._wrapper.aclose()


@overload
def create_client(
    lusid_url: str,
    token: str,
    *,
    is_async: Literal[False] = False,
) -> httpx.Client: ...


@overload
def create_client(
    lusid_url: str,
    token: str,
    *,
    is_async: Literal[True],
) -> httpx.AsyncClient: ...


def create_client(
    lusid_url: str,
    token: str,
    *,
    is_async: bool = False,
):
    if not token or not lusid_url:
        raise RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set")

    headers = {
        "authorization": f"Bearer {token}",
        "accept": "application/json",
        "content-type": "application/json",
        "X-LUSID-Application": "fbnconfig",
    }

    if is_async:
        return httpx.AsyncClient(
            base_url=lusid_url,
            timeout=60,
            headers=headers,
            event_hooks={"response": [response_hook_async]},
            transport=AsyncRetryTransport(),
        )

    return httpx.Client(
        base_url=lusid_url,
        timeout=60,
        headers=headers,
        event_hooks={"response": [response_hook]},
        transport=RetryTransport(),
    )



import pathlib
from types import SimpleNamespace
from unittest import mock

import httpx
import pytest
from httpx import HTTPStatusError
from polyfactory.factories.pydantic_factory import ModelFactory
from pydantic import ValidationError

from fbnconfig import recipe

TEST_BASE = "https://foo.lusid.com"


class RecipeFactory(ModelFactory[recipe.RecipeResource]): ...


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeRecipeResource:
    base_url = TEST_BASE
    recipe_url = "/api/api/recipes"
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    @pytest.fixture
    def configuration_recipe(self):
        return {
            "scope": "sasd",
            "code": "Abcd",
            "market": {
                "marketRules": [
                    {
                        "key": "Quote.ClientInternal.*",
                        "supplier": "Lusid",
                        "dataScope": "luminesce-examples",
                        "quoteType": "Price",
                        "field": "mid",
                        "quoteInterval": "100D.0D",
                        "priceSource": "",
                        "sourceSystem": "Lusid",
                    },
                    {
                        "key": "FX.*.*",
                        "supplier": "Lusid",
                        "dataScope": "luminesce-examples",
                        "quoteType": "Rate",
                        "field": "mid",
                        "quoteInterval": "100D.0D",
                        "priceSource": "",
                        "sourceSystem": "Lusid",
                    },
                ],
                "suppliers": {
                    "Commodity": "Client",
                    "Credit": "Client",
                    "Equity": "Client",
                    "Fx": "Client",
                    "Rates": "Client",
                },
                "options": {
                    "defaultSupplier": "Lusid",
                    "defaultInstrumentCodeType": "ClientInternal",
                    "defaultScope": "luminesce-examples",
                    "attemptToInferMissingFx": True,
                    "calendarScope": "CoppClarkHolidayCalendars",
                    "conventionScope": "Conventions",
                },
                "specificRules": [],
                "groupedMarketRules": [],
            },
            "pricing": {
                "modelRules": [
                    {
                        "supplier": "Lusid",
                        "modelName": "SimpleStatic",
                        "instrumentType": "Bond",
                        "parameters": "",
                        "modelOptions": {"modelOptionsType": "EmptyModelOptions"},
                        "instrumentId": "",
                        "addressKeyFilters": [],
                    }
                ],
                "modelChoice": {},
                "options": {
                    "modelSelection": {"library": "Lusid", "model": "SimpleStatic"},
                    "useInstrumentTypeToDeterminePricer": False,
                    "allowAnyInstrumentsWithSecUidToPriceOffLookup": False,
                    "allowPartiallySuccessfulEvaluation": False,
                    "produceSeparateResultForLinearOtcLegs": False,
                    "enableUseOfCachedUnitResults": False,
                    "windowValuationOnInstrumentStartEnd": False,
                    "removeContingentCashflowsInPaymentDiary": False,
                    "useChildSubHoldingKeysForPortfolioExpansion": False,
                    "validateDomesticAndQuoteCurrenciesAreConsistent": False,
                    "conservedQuantityForLookthroughExpansion": "PV",
                },
                "resultDataRules": [],
            },
            "aggregation": {
                "options": {
                    "useAnsiLikeSyntax": False,
                    "allowPartialEntitlementSuccess": False,
                    "applyIso4217Rounding": False,
                }
            },
            "holding": {"taxLotLevelHoldings": True},
        }

    @pytest.fixture
    def content_hash(self):
        return "bd3c111146dce83acf358b05351ce1965cc43e87fa88bc0f9634690c372cbc20"

    def test_read_recipe(self, respx_mock):
        recipe_model = {"scope": "a", "code": "b", "market": "something"}
        source_recipe = RecipeFactory.build(recipe=recipe_model)
        remote_recipe = recipe_model | {"somedefaultvalue": "abcs"}
        respx_mock.get(f"{self.recipe_url}/{source_recipe.scope}/{source_recipe.code}").mock(
            return_value=httpx.Response(200, json={"value": remote_recipe})
        )

        old_state = SimpleNamespace(
            id="abc",
            scope=source_recipe.scope,
            code=source_recipe.code,
            local_version="somehash",
            remote_version="someotherHash")

        response = source_recipe.read(self.client, old_state=old_state)
        remote_recipe.pop("scope")
        remote_recipe.pop("code")
        assert response == remote_recipe

    def test_read_recipe_not_found(self, respx_mock, configuration_recipe):
        sut = RecipeFactory.build(recipe=configuration_recipe)
        respx_mock.get(f"{self.recipe_url}/{sut.scope}/{sut.code}").mock(return_value=httpx.Response(404))

        with pytest.raises(HTTPStatusError):
            sut.read(
                self.client,
                old_state=SimpleNamespace(
                    id="abc",
                    scope=sut.scope,
                    code=sut.code,
                    local_version="somehash",
                    remote_version="someotherHash",
                ),
            )

    def test_create_recipe(self, respx_mock, configuration_recipe, content_hash):
        remote_recipe = configuration_recipe | {"somedefault": "something"}
        respx_mock.post(self.recipe_url).mock(
            return_value=httpx.Response(200, json={"value": "2000-01-01T00:00:00+00:00"})
        )
        respx_mock.get(f"{self.recipe_url}/sc1/cd1", params={"asAt": "2000-01-01T00:00:00+00:00"}).mock(
            return_value=httpx.Response(200, json={"value": remote_recipe})
        )

        sut = recipe.RecipeResource(
            id="recipe1", scope="sc1", code="cd1", recipe=configuration_recipe
        )
        state = sut.create(client=self.client)

        assert state == {
            "scope": "sc1",
            "code": "cd1",
            "source_version": content_hash,
            "remote_version": "edd788fdb0a3f48ad5d90becd031f2e216d4763f9624c91632b568c3d0b7ffdb"
        }

    def test_delete_recipe(self, respx_mock, configuration_recipe, content_hash):
        respx_mock.delete("/api/api/recipes/sc1/cd1").mock(return_value=httpx.Response(200))

        sut = recipe.RecipeResource(
            id="recipe1", scope="sc1", code="cd1", recipe=configuration_recipe
        )
        old_state = SimpleNamespace(scope="sc1", code="cd1")
        sut.delete(client=self.client, old_state=old_state)

    def test_delete_recipe_not_found(self, respx_mock, configuration_recipe, content_hash):
        respx_mock.delete("/api/api/recipes/sc1/cd1").mock(return_value=httpx.Response(404))

        sut = recipe.RecipeResource(
            id="recipe1", scope="sc1", code="cd1", recipe=configuration_recipe
        )
        old_state = SimpleNamespace(scope="sc1", code="cd1")
        with pytest.raises(HTTPStatusError):
            sut.delete(client=self.client, old_state=old_state)

    def test_recipe_source_content_changes(self, respx_mock, configuration_recipe, content_hash):
        remote_recipe = {"scope": "a", "code": "b", "somedefault": "value"}
        as_at = "2025-02-19T00:00:00+00:00"

        # GIVEN A recipe already exists
        sut = recipe.RecipeResource(id="recipe1", scope="sc1", code="cd1", recipe=configuration_recipe)
        respx_mock.get("/api/api/recipes/sc1/cd1").mock(side_effect=[
            httpx.Response(200, json={"value": remote_recipe}),
            httpx.Response(200, json={"value": remote_recipe} | {"somekey": "Values"})])

        # AND Update is called
        respx_mock.post("/api/api/recipes").mock(
            return_value=httpx.Response(200, json={"value": as_at})
        )
        old_state = SimpleNamespace(
            scope="sc1",
            code="cd1",
            source_version=content_hash,
            remote_version="94bcd66609372eb7974f310e906aff33e97d3ff71afa9b375de47a2fc96a7ee6"
        )

        # WHEN The source recipe changes
        if isinstance(sut.recipe, dict):
            sut.recipe.update({"heres": "somethingnew"})

        state = sut.update(client=self.client, old_state=old_state)
        assert state is not None
        assert state["source_version"] != content_hash
        assert state == {
            "scope": "sc1",
            "code": "cd1",
            "source_version": "a63bc7db459fc9ef80753fd25042bed84e213a4ffbb7792cb86ba3e08911c52b",
            "remote_version": "71cdcf3e703becaae96946eba8984aed745431f39a88605d5931f3217598ead7",
        }

        assert respx_mock.calls.last.request.url.params.get("asAt") == as_at

    def test_recipe_remote_content_changes(self, respx_mock, configuration_recipe):
        remote_recipe = {"scope": "a", "code": "b", "something new": "value"}
        as_at = "2025-02-19T00:00:00+00:00"

        # GIVEN A recipe already exists
        sut = recipe.RecipeResource(id="recipe1", scope="sc1", code="cd1", recipe=configuration_recipe)
        respx_mock.get("/api/api/recipes/sc1/cd1").mock(
            side_effect=[
                httpx.Response(200, json={"value": remote_recipe}),
                httpx.Response(200, json={"value": configuration_recipe} | {"somedefault": "Values"}),
            ]
        )
        content_hash = "a63bc7db459fc9ef80753fd25042bed84e213a4ffbb7792cb86ba3e08911c52b"
        # AND Update is called
        respx_mock.post("/api/api/recipes").mock(return_value=httpx.Response(200, json={"value": as_at}))
        old_state = SimpleNamespace(
            scope="sc1",
            code="cd1",
            source_version=content_hash,
            remote_version="94bcd66609372eb7974f310e906aff33e97d3ff71afa9b375de47a2fc96a7ee6",
        )

        # WHEN The source recipe changes
        if isinstance(sut.recipe, dict):
            sut.recipe.update({"heres": "somethingnew"})

        state = sut.update(client=self.client, old_state=old_state)
        assert state == {
            "scope": "sc1",
            "code": "cd1",
            "source_version": content_hash,
            "remote_version": "c3c325b634053f0bc4e0b0ef3b914f514c8c03dff0bd3e792a0fab21c7079053",
        }

        assert respx_mock.calls.last.request.url.params.get("asAt") == as_at

    @pytest.mark.parametrize("scope_and_code", [("sc2", "cd2"), ("sc1", "cd2"), ("sc2", "cd1")])
    def test_recipe_identifier_change(self, respx_mock, configuration_recipe, scope_and_code):
        scope, code = scope_and_code

        respx_mock.delete(f"/api/api/recipes/{scope}/{code}").mock(
            return_value=httpx.Response(200)
        )
        respx_mock.post("/api/api/recipes").mock(
            return_value=httpx.Response(200, json={"value": "2000-01-01T00:00:00+00:00"})
        )
        respx_mock.get("/api/api/recipes/sc1/cd1", params={"asAt": "2000-01-01T00:00:00+00:00"}).mock(
            return_value=httpx.Response(200, json={"value": {"scope": "a", "code": "b",
                                                             "original": "same object"}})
        )

        sut = recipe.RecipeResource(
            id="recipe1", scope="sc1", code="cd1", recipe=configuration_recipe
        )
        old_state = SimpleNamespace(
            scope=scope, code=code, source_version="something", remote_verson="somethingelse"
        )
        state = sut.update(client=self.client, old_state=old_state)
        assert state == {
            "scope": "sc1",
            "code": "cd1",
            "source_version": "bd3c111146dce83acf358b05351ce1965cc43e87fa88bc0f9634690c372cbc20",
            "remote_version": "2ae263bc3b030b776c16a689bc0d39edcb68ec6a3b1488b4c1759b5a75205333",
        }
        assert state != old_state

    def test_update_recipe_no_change(self, respx_mock, configuration_recipe, content_hash):
        remote_recipe = {"value": configuration_recipe | {"somedefault": "val"}}

        respx_mock.get("/api/api/recipes/sc1/cd1").mock(
            return_value=httpx.Response(200, json=remote_recipe)
        )

        sut = recipe.RecipeResource(
            id="recipe1", scope="sc1", code="cd1", recipe=configuration_recipe
        )
        old_state = SimpleNamespace(scope="sc1", code="cd1",
                                    source_version=content_hash,
                                    remote_version="def1eab0f0d8b5e90bedfa32b076efaeb379873e4bb0c79cf4f5f9c439eb7a5e")
        state = sut.update(client=self.client, old_state=old_state)
        assert state is None

    @staticmethod
    def test_read_file(configuration_recipe, content_hash):
        mock_open = mock.mock_open(read_data=b'{"test": "12"}')
        with mock.patch("builtins.open", mock_open) as m:
            sut = recipe.RecipeResource(
                id="from-file", recipe=pathlib.PurePath("/foo.txt"), scope="sc1", code="cd1"
            )
            assert sut.recipe == pathlib.PurePath("/foo.txt")
            assert sut._recipe_model_ == {"test": "12"}
            # not asserting content hash as we haven't ran a read yet
            m.assert_called_once_with(pathlib.PurePosixPath("/foo.txt"), "rb")

        mock_open_file_update = mock.mock_open(read_data=b'{"test": "34"}')
        with mock.patch("builtins.open", mock_open_file_update) as m:
            sut = recipe.RecipeResource(
                id="from-file", recipe=pathlib.PurePath("/foo.txt"), scope="sc1", code="cd1"
            )
            # not asserting content hash as we haven't ran a read yet
            assert sut.recipe == pathlib.PurePath("/foo.txt")
            assert sut._recipe_model_ == {"test": "34"}
            m.assert_called_once_with(pathlib.PurePosixPath("/foo.txt"), "rb")

    @staticmethod
    def test_ctor_throws_runtime_error_when_path_and_str_not_exists():
        with pytest.raises(ValidationError):
            # noinspection PyArgumentList
            recipe.RecipeResource(id="recipe1", scope="sc1", code="cd1")  # type:ignore

    @staticmethod
    def test_ctor_throws_value_error_when_path_not_absolute():
        with pytest.raises(ValueError):
            recipe.RecipeResource(
                id="recipe1", scope="sc1", code="cd1", recipe=pathlib.Path("./foo.txt")
            )

    @staticmethod
    def test_deps():
        sut = recipe.RecipeResource(id="recipe1", scope="sc1", code="cd1", recipe={})
        assert sut.deps() == []

    def test_dump(self):
        # given a recipe resource
        recipe_data = {
            "market": {
                "marketRules": [
                    {
                        "key": "Quote.ClientInternal.*",
                        "supplier": "Lusid",
                        "dataScope": "test-scope",
                        "quoteType": "Price",
                        "field": "mid",
                        "quoteInterval": "100D.0D",
                        "priceSource": "",
                        "sourceSystem": "Lusid",
                    }
                ]
            }
        }
        sut = recipe.RecipeResource(
            id="test-recipe",
            scope="test-scope",
            code="test-code",
            recipe=recipe_data
        )
        # when we dump it
        result = sut.model_dump(
            mode="json", by_alias=True, round_trip=True, exclude_none=True, context={"style": "dump"}
        )
        # then it's correctly serialized
        assert result == {
            "scope": "test-scope",
            "code": "test-code",
            "recipe": recipe_data
        }

    def test_undump(self):
        # given dump data
        recipe_data = {
            "market": {
                "marketRules": [
                    {
                        "key": "FX.*.*",
                        "supplier": "Lusid",
                        "dataScope": "test-scope",
                        "quoteType": "Rate",
                        "field": "mid",
                        "quoteInterval": "100D.0D",
                        "priceSource": "",
                        "sourceSystem": "Lusid",
                    }
                ]
            }
        }
        data = {
            "scope": "test-scope",
            "code": "test-code",
            "recipe": recipe_data
        }
        # when we undump it
        result = recipe.RecipeResource.model_validate(
            data, context={"style": "undump", "id": "test-recipe"}
        )
        # then it's correctly populated
        assert result.id == "test-recipe"
        assert result.scope == "test-scope"
        assert result.code == "test-code"
        assert result.recipe == recipe_data
        assert result._recipe_model_ == recipe_data

    def test_from_api(self, testdata_loader):
        # GIVEN a Recipe resource
        input_data = testdata_loader(
            "recipe_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "recipe_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # WHEN the from_api method is called for that resource
        result = recipe.RecipeResource.from_api(input_data)
        # THEN the response is as expected
        assert result == expected_result
        # AND model is validated successfully
        recipe.RecipeResource.model_validate(result)

    @pytest.mark.asyncio
    async def test_list(self, respx_mock, testdata_loader):
        # GIVEN a Recipe resource
        input_data = testdata_loader(
            "recipe_resource",
            "list_mock_data"
        )
        expected_result = testdata_loader(
            "recipe_resource",
            "list_expected_result"
        )
        # AND the resource exists in LUSID
        path_params = {}
        url = recipe.RecipeResource.list_endpoint().format(**path_params)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the list method is called for that resource
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            result = await recipe.RecipeResource.list(client, path_params=path_params)
        # THEN the response is as expected
        assert result == expected_result

    @pytest.mark.asyncio
    async def test_fetch(self, respx_mock, testdata_loader):
        # GIVEN a Recipe resource
        input_data = testdata_loader(
            "recipe_resource",
            "list_mock_data"
        )["values"][0]
        expected_result = testdata_loader(
            "recipe_resource",
            "list_expected_result"
        )["values"][0]["value"]
        # AND the resource exists in LUSID
        keys = {
            "scope": input_data["value"]["scope"],
            "code": input_data["value"]["code"],
        }
        url = recipe.RecipeResource.fetch_endpoint().format(**keys)
        respx_mock.get(url).mock(
            return_value=httpx.Response(200, json=input_data)
        )
        # WHEN the fetch method is called for that resource
        async with httpx.AsyncClient(base_url=TEST_BASE) as client:
            result = await recipe.RecipeResource.fetch(client, keys=keys)
        # THEN the response is as expected
        assert result == expected_result
        # AND the correct request was made to LUSID
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == url
        assert request.url.query == b""



import functools
from typing import Any

import httpx


class QueryError(RuntimeError):  # has to be runtime for backward compat
    def __init__(self, status_str: str, status_resp: Any):
        self.status_str = status_str
        self.status_resp = status_resp
        super().__init__(self._get_message())

    def _get_message(self) -> str:
        info = []
        summary = ""
        if self.status_resp:
            if self.status_resp.get("feedback"):
                info = [
                    f"{k}: {v}"
                    for k, v in self.status_resp["feedback"][-1].items()
                    if k in ["executionId", "sessionId"]
                ]
            if self.status_resp.get("progress", None):
                summary = self.status_resp["progress"].split("Dependency & Execution details:")[-1]
                summary = (" " * 8).join(summary.splitlines(True))
        return "\n".join([f"Error, query {self.status_str}"] + info + [summary])

    def __str__(self) -> str:
        return self._get_message()


class ApiError(Exception):
    pass


class ApiStatusError(Exception):
    def __init__(self, http_exception: httpx.HTTPStatusError):
        self.http_exception = http_exception
        super().__init__(self._get_message())

    def _get_message(self) -> str:
        exc = self.http_exception
        # build the detail
        status = exc.response.status_code
        base_msg = " ".join(
            [
                f"The server responded with {exc.response.status_code}",
                f"while requesting {exc.request.method} {exc.request.url!r}",
            ]
        )
        if exc.response.is_server_error:
            detail = [exc.response.text]
        elif status == 401:
            detail = [
                "Unauthorized: The user making the request could not be identified.",
                "This may indicate that the provided access token is invalid or expired."
            ]
        elif status == 403:
            detail = [
                "Access Denied",
                "User does not have permission to perform the requested action"
            ]
        else:
            detail = []
        # build the subdetail if it's a finbourne problem detail
        if "application/problem+json" in exc.response.headers.get("content-type", "").split(";"):
            problem = exc.response.json()
            subdetail = [
                f"{k}: {v}"
                 for k, v in problem.items()
                 if k in ["title", "detail", "instance"]
            ]
            if problem.get("errors", None):
                subdetail = subdetail + ["errors: " + str(problem["errors"].get("Name", ""))]
        else:
            subdetail = [f"detail {exc.response.text}"]
        return "\n".join([base_msg] + detail + subdetail)

    def __str__(self) -> str:
        return self._get_message()


class NotFoundError(ApiError):
    def __init__(self, entity_name: str):
        self.message = f"{entity_name} not found".capitalize()
        super().__init__(self.message)


def api_exception_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        # deal with errors that do not send a response
        except httpx.RequestError as exc:
            msg = " ".join([
                f"An error occurred while requesting {exc.request.url!r} ",
                f"before we got a response - {exc}"
            ])
            raise ApiError(msg) from exc
        # errors from the server with a response
        except httpx.HTTPStatusError as exc:
            raise ApiStatusError(exc) from exc

    return wrapper



import httpx
import respx


def test_httpx_client_with_retry_hook():
    router = respx.Router()
    router.get("http://fakeexample.com/200").mock(
        side_effect=[
            httpx.Response(429, json={"id": 1234}, headers={"Retry-after": "0.1"}),
            httpx.Response(200, json={"id": 23456}),
        ]
    )
    mock_transport = httpx.MockTransport(router.handler)

    client = httpx.Client(base_url="http://fakeexample.com", transport=mock_transport)
    retries = 0
    sleeps = []

    def sleeper(t):
        nonlocal sleeps
        sleeps.append(t)

    def response_hook(response):
        nonlocal retries
        if response.status_code == 429:
            retry_header = response.headers.get("retry-after", None)
            delay = float(retry_header) if retry_header else 1
            sleeper(delay)
            retries = 1
            retry = client.send(response.request)
            response.next_request = retry.next_request
            response.status_code = retry.status_code
            response.headers = retry.headers
            response._content = retry._content

    client.event_hooks["response"] = [response_hook]
    res = client.request("get", "/200")
    assert sleeps == [0.1]
    assert retries == 1
    assert res.status_code == 200
    assert res.json() == {"id": 23456}
    sent = router.calls.last.request
    assert sent.method == "GET"
    assert sent.url == "http://fakeexample.com/200"


def test_httpx_client_with_raise_for_status():
    router = respx.Router()
    router.post("http://fakeexample.com/test").mock(
        side_effect=[
            httpx.Response(400, json={"not found": 1234}),
            httpx.Response(200, json={"id": 23456}),
        ]
    )

    def response_hook(response):
        response.raise_for_status()

    mock_transport = httpx.MockTransport(router.handler)

    client = httpx.Client(
        base_url="http://fakeexample.com",
        transport=mock_transport,
        event_hooks={"response": [response_hook]},
    )

    try:
        client.request("post", "/test")
    except httpx.HTTPStatusError as error:
        assert error.response.status_code == httpx.codes.BAD_REQUEST
        assert error.response.json()["not found"] == 1234

    sent = router.calls.last.request
    assert sent.method == "POST"
    assert sent.url.path == "/test"

    res = client.request("post", "/test")
    assert res.status_code == 200
    assert res.json() == {"id": 23456}
    sent = router.calls.last.request
    assert sent.method == "POST"
    assert sent.url.path == "/test"



from fbnconfig import Deployment, customentity, datatype, property

"""
An example configuration for a custom entity.
The script configures the following entities:
- Entity Type

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01750/en-us
"""


def configure(env):
    # entity type
    ce_type = customentity.EntityTypeResource(
        id="ce1",
        entity_type_name="entity-type-name",
        display_name="Example Custom Entity",
        description="An example custom entity",
        field_schema=[
            customentity.FieldDefinition(
                name="Field1",
                lifetime=customentity.LifeTime.PERPETUAL,
                type=customentity.FieldType.STRING,
                collection_type=customentity.CollectionType.SINGLE,
                required=True,
            ),
            customentity.FieldDefinition(
                name="Field2",
                lifetime=customentity.LifeTime.TIMEVARIANT,
                type=customentity.FieldType.STRING,
                required=True,
            ),
        ],
    )
    # data type
    string = datatype.DataTypeRef(
        id="systemstring",
        scope="system",
        code="string",
    )
    # identifier
    identifier_type = property.DefinitionResource(
        id="exid",
        domain=property.Domain.CustomEntity,
        scope="ce-example",
        code="ce-example-id",
        display_name="Example Custom Entity ID",
        data_type_id=string,
        life_time=property.LifeTime.Perpetual,
        constraint_style=property.ConstraintStyle.Identifier,
    )
    property_definition = property.DefinitionResource(
        id="propid",
        domain=property.Domain.CustomEntity,
        scope="ce-example",
        code="ce-example-prop",
        display_name="Example Property on a CE",
        data_type_id=string,
        life_time=property.LifeTime.Perpetual,
        constraint_style=property.ConstraintStyle.Property,
    )
    # entity instance with properties
    ce_instance = customentity.EntityResource(
        id="ce1-instance",
        entity_type=ce_type,
        description="An example custom entity instance",
        display_name="Example Instance",
        identifiers=[
            customentity.EntityIdentifier(
                identifier_type=identifier_type,
                identifier_value="ce-exmaple-one",
            )
        ],
        fields=[
            customentity.EntityField(
                name="Field1",
                value="Example Value 1",
            ),
            customentity.EntityField(
                name="Field2",
                value="Example Value 2",
            )
        ],
        properties=[customentity.PropertyValue(
            property_key=property_definition,
            label_value="banjo",
        )]
    )
    return Deployment("custom_entity_example", [ce_type, ce_instance, identifier_type])



import os
from types import SimpleNamespace

import pytest
from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
from fbnconfig import notifications as notif
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise RuntimeError(
        "Both FBN_ACCESS_TOKEN and LUSID_ENV variables need "
        "to be set to run integration tests"
    )


lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]


@fixture(scope="module")
def setup_deployment():
    print("Creating new subscription")
    client = fbnconfig.create_client(lusid_env, token)

    # Deletes test subscription if it exsits
    try:
        client.delete("/notification/api/subscriptions/testScope/testCode")
    except Exception:
        pass

    # Creates new subscription to test notifications on
    desired = {
        "id": {
            "scope": "testScope",
            "code": "testCode"
        },
        "displayName": "TestDisplayName",
        "description": "TestDescription",
        "status": "Active",
        "matchingPattern": {
            "eventType": "Manual",
            "filter": "Body.Message eq 'TestMessage'"
        },
    }
    client.post("/notification/api/subscriptions", json=desired)

    deployment_name = gen("notifications")
    print(f"\nRunning for deployment {deployment_name}...")

    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    client.delete("/notification/api/subscriptions/testScope/testCode")


def configure(env):
    """Create a test email notification."""
    deployment_name = getattr(env, "name", "notification")

    email_type = notif.EmailNotificationType(
        subject="Test Event: {{body.subject}}",
        plain_text_body="Event received: {{body.message}}",
        html_body="<p>Event received: <strong>{{body.message}}</strong></p>",
        email_address_to=["test@finbourne.com"],
        email_address_cc=["testcc@finbourne.com"]
    )

    subscription_ref = notif.SubscriptionRef(
        id="subscription_ref",
        scope="testScope",
        code="testCode"
    )

    email_notification = notif.NotificationResource(
        id="testEmailNotification",
        subscription=subscription_ref,
        notification_id=f"emailNotif001-{deployment_name}",
        display_name="Test Email Notification",
        description="Testing email notification creation",
        notification_type=email_type
    )

    matching_pattern = notif.MatchingPattern(
        event_type="Manual",
        filter="Body.Message eq 'Test'"
    )

    sub_resource = notif.SubscriptionResource(
        id="SubExampleId",
        scope=f"testScopeSub-{deployment_name}",
        code="testCodeSub",
        display_name="Example display name",
        description="Example description",
        status=notif.SubscriptionStatus.ACTIVE,
        matching_pattern=matching_pattern
    )

    sms_type = notif.SmsNotificationType(
        body="Example_body",
        recipients=["+47000000000"]
    )

    sms_notification = notif.NotificationResource(
        id="testSmsNotification",
        subscription=sub_resource,
        notification_id=f"smsNotif001-{deployment_name}",
        display_name="Test sms Notification",
        description="Testing sms notification creation",
        notification_type=sms_type
    )

    return fbnconfig.Deployment(deployment_name, [email_notification, sms_notification])


def test_teardown(setup_deployment):
    deployment_name = setup_deployment.name
    client = fbnconfig.create_client(lusid_env, token)
    # setup deployment
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)

    # check they exists
    client.request("get", "/notification/api/subscriptions/testScope/testCode/"
        f"notifications/emailNotif001-{deployment_name}")

    client.get(f"/notification/api/subscriptions/testScopeSub-{deployment_name}/testCodeSub/"
        f"notifications/smsNotif001-{deployment_name}")

    # Tear it down
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)

    # check they were deleted
    try:
        client.request("get", "/notification/api/subscriptions/testScope/testCode/"
        f"notifications/emailNotif001-{deployment_name}")
    except HTTPStatusError as error:
        assert error.response.status_code == 404

    try:
        client.get(f"/notification/api/subscriptions/testScopeSub-{deployment_name}/testCodeSub/"
        f"notifications/smsNotif001-{deployment_name}")
    except HTTPStatusError as error:
        assert error.response.status_code == 404


def test_create_email_notification(setup_deployment):
    deployment_name = setup_deployment.name
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)

    # checks email and sms notifications are created
    email_response = client.get("/notification/api/subscriptions/testScope/testCode/"
    f"notifications/emailNotif001-{deployment_name}")

    sms_response = client.get(f"/notification/api/subscriptions/"
    f"testScopeSub-{deployment_name}/testCodeSub/notifications/smsNotif001-{deployment_name}")

    # assert codes
    assert email_response.status_code == 200
    assert sms_response.status_code == 200


def test_update_no_changes(setup_deployment):
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    update = fbnconfig.deployex(configure(setup_deployment), lusid_env, token)

    email_change = [a.change for a in update
                    if a.type == "NotificationResource" and a.id == "testEmailNotification"]
    sms_change = [a.change for a in update
                  if a.type == "NotificationResource" and a.id == "testSmsNotification"]
    assert email_change == ["nochange"]
    assert sms_change == ["nochange"]


@pytest.mark.asyncio
async def test_notification_list(setup_deployment):
    # GIVEN Notification resource is deployed
    deployment_name = setup_deployment.name
    subscription_scope = "testScope"
    subscription_code = "testCode"
    notification_id = f"emailNotif001-{deployment_name}"
    # WHEN calling list with filter for the deployment name
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    result = await notif.NotificationResource.list(
        client=client,
        query_params={},
        path_params={
            "subscriptionScope": subscription_scope,
            "subscriptionCode": subscription_code,
        }
    )
    notification = next(
        (i for i in result["values"] if i["keys"]["notificationId"] == notification_id)
    )
    # THEN the result is as expected
    expected_notification = {
        "keys": {
            "subscriptionScope": subscription_scope,
            "subscriptionCode": subscription_code,
            "notificationId": notification_id,
        },
        "value": {
            "id": f"notification_{subscription_scope}_{subscription_code}_{notification_id}".lower(),
            "notificationId": notification_id,
            "description": "Testing email notification creation",
            "displayName": "Test Email Notification",
            "subscription": {
                "$ref": f"subscription_{subscription_scope}_{subscription_code}".lower()
            },
            "notificationType": {
                "type": "Email",
                "subject": "Test Event: {{body.subject}}",
                "plainTextBody": "Event received: {{body.message}}",
                "htmlBody": "<p>Event received: <strong>{{body.message}}</strong></p>",
                "emailAddressBcc": [],
                "emailAddressCc": ["testcc@finbourne.com"],
                "emailAddressTo": ["test@finbourne.com"],
            }
        }
    }
    assert notification == expected_notification


@pytest.mark.asyncio
async def test_notification_fetch(setup_deployment):
    # GIVEN Notification resource is deployed
    deployment_name = setup_deployment.name
    subscription_scope = "testScope"
    subscription_code = "testCode"
    notification_id = f"emailNotif001-{deployment_name}"
    client = fbnconfig.create_client(lusid_env, token, is_async=True)
    # WHEN the fetch method is called for that resource
    response = await notif.NotificationResource.fetch(
        client=client,
        keys={
            "subscriptionScope": subscription_scope,
            "subscriptionCode": subscription_code,
            "notificationId": notification_id,
        }
    )
    # THEN the response is as expected
    expected_response = {
        "id": f"notification_{subscription_scope}_{subscription_code}_{notification_id}".lower(),
        "notificationId": notification_id,
        "description": "Testing email notification creation",
        "displayName": "Test Email Notification",
        "subscription": {
            "$ref": f"subscription_{subscription_scope}_{subscription_code}".lower()
        },
        "notificationType": {
            "type": "Email",
            "subject": "Test Event: {{body.subject}}",
            "plainTextBody": "Event received: {{body.message}}",
            "htmlBody": "<p>Event received: <strong>{{body.message}}</strong></p>",
            "emailAddressBcc": [],
            "emailAddressCc": ["testcc@finbourne.com"],
            "emailAddressTo": ["test@finbourne.com"],
        }
    }
    assert response == expected_response
